# Copyright European Organization for Nuclear Research (CERN) since 2012
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import TYPE_CHECKING

from sqlalchemy import and_, delete, select
from sqlalchemy.exc import NoResultFound

from rucio.common.exception import CounterNotFound
from rucio.db.sqla import filter_thread_work, models
from rucio.db.sqla.session import read_session, transactional_session

if TYPE_CHECKING:
    from sqlalchemy.orm import Session


@transactional_session
def add_counter(rse_id, *, session: "Session"):
    """
    Creates the specified counter for a rse_id.

    :param rse_id:  The id of the RSE.
    :param session: The database session in use.
    """
    models.RSEUsage(rse_id=rse_id, source='rucio', files=0, used=0).\
        save(session=session)


@transactional_session
def increase(rse_id, files, bytes_, *, session: "Session"):
    """
    Increments the specified counter by the specified amount.

    :param rse_id:  The id of the RSE.
    :param files:   The number of added files.
    :param bytes_:   The number of added bytes.
    :param session: The database session in use.
    """
    models.UpdatedRSECounter(rse_id=rse_id, files=files, bytes=bytes_).\
        save(session=session)


@transactional_session
def decrease(rse_id, files, bytes_, *, session: "Session"):
    """
    Decreases the specified counter by the specified amount.

    :param rse_id:  The id of the RSE.
    :param files:   The number of removed files.
    :param bytes_:   The number of removed bytes.
    :param session: The database session in use.
    """
    return increase(rse_id=rse_id, files=-files, bytes_=-bytes_, session=session)


@transactional_session
def del_counter(rse_id, *, session: "Session"):
    """
    Delete specified counter.

    :param rse_id:  The id of the RSE.
    :param session: The database session in use.
    """

    stmt = delete(
        models.RSEUsage
    ).where(
        and_(models.RSEUsage.rse_id == rse_id,
             models.RSEUsage.source == 'rucio')
    ).execution_options(
        synchronize_session=False
    )
    session.execute(stmt)


@read_session
def get_counter(rse_id, *, session: "Session"):
    """
    Returns current values of the specified counter or raises CounterNotFound if the counter does not exist.

    :param rse_id:           The id of the RSE.
    :param session:          The database session in use.
    :returns:                A dictionary with total and bytes.
    :raises CounterNotFound: If the counter does not exist.
    :returns:                A dictionary with total and bytes.
    """

    try:
        stmt = select(
            models.RSEUsage.used,
            models.RSEUsage.files,
            models.RSEUsage.updated_at
        ).where(
            and_(models.RSEUsage.rse_id == rse_id,
                 models.RSEUsage.source == 'rucio')
        )
        usage_bytes, usage_files, usage_updated_at = session.execute(stmt).one()
        return {
            'bytes': usage_bytes,
            'files': usage_files,
            'updated_at': usage_updated_at
        }
    except NoResultFound:
        raise CounterNotFound()


@read_session
def get_updated_rse_counters(total_workers, worker_number, *, session: "Session"):
    """
    Get updated rse_counters.

    :param total_workers:      Number of total workers.
    :param worker_number:      id of the executing worker.
    :param session:            Database session in use.
    :returns:                  List of rse_ids whose rse_counters need to be updated.
    """
    stmt = select(
        models.UpdatedRSECounter.rse_id
    ).distinct(
    )

    stmt = filter_thread_work(session=session, query=stmt, total_threads=total_workers, thread_id=worker_number, hash_variable='rse_id')
    return session.execute(stmt).scalars().all()


@transactional_session
def update_rse_counter(rse_id, *, session: "Session"):
    """
    Read the updated_rse_counters and update the rse_counter.

    :param rse_id:   The rse_id to update.
    :param session:  Database session in use.
    """

    stmt = select(
        models.UpdatedRSECounter
    ).where(
        models.UpdatedRSECounter.rse_id == rse_id
    )
    updated_rse_counters = session.execute(stmt).scalars().all()
    sum_bytes = sum([updated_rse_counter.bytes for updated_rse_counter in updated_rse_counters])
    sum_files = sum([updated_rse_counter.files for updated_rse_counter in updated_rse_counters])

    try:
        stmt = select(
            models.RSEUsage
        ).where(
            and_(models.RSEUsage.rse_id == rse_id,
                 models.RSEUsage.source == 'rucio')
        )
        rse_counter = session.execute(stmt).scalar_one()
        rse_counter.used = (rse_counter.used or 0) + sum_bytes
        rse_counter.files = (rse_counter.files or 0) + sum_files
    except NoResultFound:
        models.RSEUsage(rse_id=rse_id,
                        used=sum_bytes,
                        files=sum_files,
                        source='rucio').save(session=session)

    for update in updated_rse_counters:
        update.delete(flush=False, session=session)


@transactional_session
def fill_rse_counter_history_table(*, session: "Session"):
    """
    Fill the RSE usage history table with the current usage.

    :param session: Database session in use.
    """
    stmt = select(
        models.RSEUsage
    )
    for usage in session.execute(stmt).scalars().all():
        models.RSEUsageHistory(rse_id=usage['rse_id'], used=usage['used'], files=usage['files'], source=usage['source']).save(session=session)



# Copyright European Organization for Nuclear Research (CERN) since 2012
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import os
import random
import string
from collections import namedtuple
from logging import getLogger
from typing import TYPE_CHECKING

import pytest
from sqlalchemy import func, select

import rucio.gateway.rule
from rucio.client.ruleclient import RuleClient
from rucio.common.config import config_get_bool
from rucio.common.constants import RseAttr
from rucio.common.exception import (
    AccessDenied,
    DuplicateRule,
    InputValidationError,
    InsufficientAccountLimit,
    InvalidSourceReplicaExpression,
    InvalidValueForKey,
    ManualRuleApprovalBlocked,
    RSEOverQuota,
    RSEWriteBlocked,
    RucioException,
    RuleNotFound,
    RuleReplaceFailed,
    UnsupportedOperation,
)
from rucio.common.schema import get_schema_value
from rucio.common.types import InternalAccount, InternalScope
from rucio.common.utils import generate_uuid as uuid
from rucio.core.account import add_account_attribute, get_account, get_usage
from rucio.core.account_limit import set_global_account_limit, set_local_account_limit
from rucio.core.did import add_did, attach_dids, set_status
from rucio.core.lock import get_dataset_locks, get_replica_locks, successful_transfer
from rucio.core.replica import add_replica, get_replica
from rucio.core.request import get_request_by_did
from rucio.core.rse import add_rse, add_rse_attribute, del_rse, del_rse_attribute, get_rse_id, set_rse_limits, update_rse
from rucio.core.rse_counter import get_counter as get_rse_counter
from rucio.core.rule import add_rule, add_rules, delete_rule, get_rule, list_rules, move_rule, reduce_rule, update_rule
from rucio.core.scope import add_scope
from rucio.daemons.abacus.account import account_update
from rucio.daemons.abacus.rse import rse_update
from rucio.daemons.judge.evaluator import re_evaluator
from rucio.db.sqla import models
from rucio.db.sqla.constants import OBSOLETE, DIDType, LockState, RuleState
from rucio.db.sqla.session import transactional_session
from rucio.gateway.account import add_account
from rucio.tests.common import account_name_generator, did_name_generator, rse_name_generator
from rucio.tests.common_server import get_vo

if TYPE_CHECKING:
    from .temp_factories import TemporaryDidFactory, TemporaryRSEFactory

LOG = getLogger(__name__)
RSE_namedtuple = namedtuple('RSE_namedtuple', ['name', 'id'])


def create_files(nrfiles, scope, rse_id, bytes_=1):
    """
    Creates a number of test files and add replicas to rse

    :param nrfiles:  Number of files to create
    :param scope:    Scope to create the files in
    :param rse_id:   RSE to add the replica to
    :param bytes_:    Bytes of each file
    :returns:        List of dict
    """
    if config_get_bool('common', 'multi_vo', raise_exception=False, default=False):
        vo = {'vo': get_vo()}
    else:
        vo = {}

    files = []
    jdoe = InternalAccount('jdoe', **vo)
    for i in range(nrfiles):
        file = did_name_generator('file')
        if isinstance(rse_id, list):
            for r in rse_id:
                add_replica(rse_id=r, scope=scope, name=file, bytes_=bytes_, account=jdoe)
        else:
            add_replica(rse_id=rse_id, scope=scope, name=file, bytes_=bytes_, account=jdoe)
        files.append({'scope': scope, 'name': file, 'bytes': bytes_})
    return files


def tag_generator(size=8, chars=string.ascii_uppercase):
    return ''.join(random.choice(chars) for x in range(size))


@transactional_session
def check_dataset_ok_callback(scope, name, rse, rse_id, rule_id, *, session=None):
    message = {'scope': scope.external,
               'name': name,
               'rse': rse,
               'rse_id': rse_id,
               'rule_id': rule_id}
    if scope.vo != 'def':
        message['vo'] = scope.vo

    stmt = select(
        func.count(models.Message.id)
    ).select_from(
        models.Message
    ).where(
        models.Message.payload == json.dumps(message)
    )
    return session.execute(stmt).scalar() > 0


@transactional_session
def check_rule_progress_callback(scope, name, progress, rule_id, *, session=None):
    message = {'scope': scope.external,
               'name': name,
               'rule_id': rule_id,
               'progress': progress}
    if scope.vo != 'def':
        message['vo'] = scope.vo

    stmt = select(
        models.Message.id
    ).where(
        models.Message.payload == json.dumps(message)
    )
    return session.execute(stmt).scalar_one_or_none() is not None


@pytest.fixture
def setup_class(request, vo, root_account, jdoe_account, rse_factory):
    cls = request.cls

    # Add test RSE
    cls.rse1, cls.rse1_id = rse_factory.make_mock_rse()
    cls.rse2, cls.rse2_id = rse_factory.make_mock_rse()
    cls.rse3, cls.rse3_id = rse_factory.make_mock_rse()
    cls.rse4, cls.rse4_id = rse_factory.make_mock_rse()

    # Add Tags
    cls.T1 = tag_generator()
    cls.T2 = tag_generator()
    add_rse_attribute(cls.rse1_id, cls.T1, True)
    add_rse_attribute(cls.rse2_id, cls.T1, True)
    add_rse_attribute(cls.rse3_id, cls.T2, True)
    add_rse_attribute(cls.rse4_id, cls.T1, True)

    # Add fake weights
    add_rse_attribute(cls.rse1_id, "fakeweight", 10)
    add_rse_attribute(cls.rse2_id, "fakeweight", 0)
    add_rse_attribute(cls.rse3_id, "fakeweight", 0)
    add_rse_attribute(cls.rse4_id, "fakeweight", 0)

    set_local_account_limit(jdoe_account, cls.rse1_id, -1)
    set_local_account_limit(jdoe_account, cls.rse2_id, -1)
    set_local_account_limit(jdoe_account, cls.rse3_id, -1)
    set_local_account_limit(jdoe_account, cls.rse4_id, -1)


@pytest.mark.usefixtures('setup_class')
class TestCore:

    @pytest.mark.parametrize("caches_mock", [{"caches_to_mock": [
        'rucio.core.rse_expression_parser.REGION'
    ]}], indirect=True)
    @pytest.mark.parametrize(
        "delete_rse", [True, False]
    )
    def test_move_rule_invalid_source_replica_expression(
        self,
        caches_mock,
        delete_rse: bool,
        did_factory: "TemporaryDidFactory",
        rse_factory: "TemporaryRSEFactory",
        root_account: "InternalAccount"
    ):
        """
        REPLICATION RULE (CORE): Move a rule using an invalid source replica expression.

        Parametrised so that there are two different ways for the source replica
        expression to become invalid:
        1) by deleting the source RSE
        2) by referring to the source RSE via an RSE attribute and then deleting
            said RSE attribute from the source RSE
        in both cases the source replica expression is invalidated since it
        evaluates to an empty set.
        """

        from hashlib import sha256

        from dogpile.cache.api import NO_VALUE

        # create RSEs A, B, C (structured into namedtuples)
        rse_a, rse_b, rse_c = (
            RSE_namedtuple(*rse_factory.make_mock_rse()) for _ in range(3)
        )
        rsekey, rseval = tag_generator(), tag_generator()
        add_rse_attribute(rse_a.id, rsekey, rseval)

        # create DID (dataset is easiest)
        dataset = did_factory.make_dataset()

        # create source replica expression
        if delete_rse:
            source_replica_expression = rse_a.name
        else:
            source_replica_expression = f"{rsekey}={rseval}"

        # add rule binding DID to RSE-B, source replica expression matches RSE-A
        rule = {
            "account": root_account,
            "copies": 1,
            "rse_expression": rse_b.name,
            "source_replica_expression": source_replica_expression
        }
        ruledict = add_rules([dataset], [rule])
        rule_id = ruledict[(dataset["scope"], dataset["name"])][0]

        # delete RSE-A (or otherwise unmatching from source replica expr)
        if delete_rse:
            del_rse(rse_a.id)
        else:
            del_rse_attribute(rse_a.id, rsekey)

        # must remove cached entries for the Source Replica Expression
        caches_mock[0].set(sha256(source_replica_expression.encode()).hexdigest(), NO_VALUE)

        # move-rule asserts correct error
        with pytest.raises(InvalidSourceReplicaExpression):
            move_rule(rule_id, rse_c.name)

    def test_add_rule_to_file_ask_approval(self, vo, mock_scope, jdoe_account):
        """ REPLICATION RULE (CORE): Add a replication rule, asking approval"""
        rse = rse_name_generator()
        rse_id = add_rse(rse, vo=vo)

        add_rse_attribute(rse_id, RseAttr.RULE_APPROVERS, jdoe_account)

        files = create_files(1, mock_scope, rse_id)
        output = add_rule(
            dids=files,
            account=jdoe_account,
            copies=1,
            rse_expression=str(rse),
            grouping='NONE',
            weight=None,
            lifetime=None,
            locked=False,
            subscription_id=None,
            ask_approval=True
        )
        assert len(output) == 1
        assert isinstance(output[0], str)

    @pytest.mark.parametrize(
        "param_type,param_group",
        [
            ("LOCALGROUPDISK", ("country", "country")),
            ("GROUPDISK", ("physgroup", "group"))
        ]
    )
    def test_create_recipients_list(
        self, param_type: str, param_group: tuple[str, str],
        vo, random_account, function_scope_prefix
    ):
        """
        REPLICATION RULE (CORE):
        Test creation of recipients list for rses of types LOCALGROUPDISK,
        LOCALGROUPTAPE, GROUPDISK.
        """

        rstring = function_scope_prefix + param_type
        rsegroup, accountgroup = param_group

        # create rse
        # rse must NOT have attr 'rule_approvers'
        rse = rse_name_generator()
        rse_id = add_rse(rse, vo=vo)
        add_rse_attribute(rse_id, "type", param_type)
        add_rse_attribute(rse_id, rsegroup, rstring)

        # create account
        account: InternalAccount = random_account
        add_account_attribute(
            account, f"{accountgroup}-{rstring}", "admin"
        )
        email = get_account(account).email

        # import and run relevant function
        from rucio.core.rule import _create_recipients_list
        out = _create_recipients_list(str(rse))

        assert out == [(email, account)]

    def test_add_rule_file_none(self, mock_scope, jdoe_account):
        """ REPLICATION RULE (CORE): Add a replication rule on a group of files, NONE Grouping"""
        files = create_files(3, mock_scope, self.rse1_id)
        add_rule(dids=files, account=jdoe_account, copies=2, rse_expression=self.T1, grouping='NONE', weight=None, lifetime=None, locked=False, subscription_id=None)

        # Check if the Locks are created properly
        t1 = {self.rse1_id, self.rse2_id, self.rse4_id}
        for file in files:
            rse_locks = {lock['rse_id'] for lock in get_replica_locks(scope=file['scope'], name=file['name'])}
            assert (len(t1.intersection(rse_locks)) > 0)
            assert self.rse3_id not in rse_locks

    def test_add_rule_dataset_none(self, mock_scope, did_factory, jdoe_account):
        """ REPLICATION RULE (CORE): Add a replication rule on a dataset, NONE Grouping"""
        files = create_files(3, mock_scope, self.rse1_id)
        dataset = did_factory.random_dataset_did()
        add_did(did_type=DIDType.DATASET, account=jdoe_account, **dataset)
        attach_dids(dids=files, account=jdoe_account, **dataset)

        # Add a first rule to the DS
        add_rule(dids=[dataset], account=jdoe_account, copies=2, rse_expression=self.T1, grouping='NONE', weight=None, lifetime=None, locked=False, subscription_id=None)

        # Add a second rule and check if the right locks are created
        add_rule(dids=[dataset], account=jdoe_account, copies=2, rse_expression='%s|%s' % (self.T1, self.T2), grouping='NONE', weight=None, lifetime=None, locked=False, subscription_id=None)

        # Check if the Locks are created properly
        t1 = {self.rse1_id, self.rse2_id, self.rse4_id}
        for file in files:
            rse_locks = {lock['rse_id'] for lock in get_replica_locks(scope=file['scope'], name=file['name'])}
            assert (len(t1.intersection(rse_locks)) == 2)
            assert self.rse3_id not in rse_locks

    def test_add_rule_duplicate(self, mock_scope, did_factory, jdoe_account):
        """ REPLICATION RULE (CORE): Add a replication rule duplicate"""
        files = create_files(3, mock_scope, self.rse1_id)
        dataset = did_factory.random_dataset_did()
        add_did(did_type=DIDType.DATASET, account=jdoe_account, **dataset)
        attach_dids(dids=files, account=jdoe_account, **dataset)

        # Add a first rule to the DS
        add_rule(dids=[dataset], account=jdoe_account, copies=2, rse_expression=self.T1, grouping='NONE', weight=None, lifetime=None, locked=False, subscription_id=None)

        # Add a second rule and check if the right locks are created
        pytest.raises(DuplicateRule, add_rule, dids=[dataset], account=jdoe_account, copies=2, rse_expression=self.T1, grouping='NONE', weight=None, lifetime=None, locked=False, subscription_id=None)

    def test_add_rules_datasets_none(self, mock_scope, did_factory, jdoe_account, root_account):
        """ REPLICATION RULE (CORE): Add replication rules to multiple datasets, NONE Grouping"""
        files1 = create_files(3, mock_scope, self.rse3_id)
        dataset1 = did_factory.random_dataset_did()
        add_did(did_type=DIDType.DATASET, account=jdoe_account, **dataset1)
        attach_dids(dids=files1, account=jdoe_account, **dataset1)

        files2 = create_files(3, mock_scope, self.rse3_id)
        dataset2 = did_factory.random_dataset_did()
        add_did(did_type=DIDType.DATASET, account=jdoe_account, **dataset2)
        attach_dids(dids=files2, account=jdoe_account, **dataset2)

        # Add the rules to both DS
        add_rules(dids=[dataset1, dataset2],
                  rules=[{'account': jdoe_account,
                          'copies': 1,
                          'rse_expression': self.T1,
                          'grouping': 'NONE',
                          'weight': None,
                          'lifetime': None,
                          'locked': False,
                          'subscription_id': None},
                         {'account': root_account,
                          'copies': 1,
                          'rse_expression': self.T1,
                          'grouping': 'NONE',
                          'weight': 'fakeweight',
                          'lifetime': None,
                          'locked': False,
                          'subscription_id': None}])

        # Check if the Locks are created properly
        for file in files1:
            rse_locks = [lock['rse_id'] for lock in get_replica_locks(scope=file['scope'], name=file['name'])]
            assert (rse_locks[0] == rse_locks[1])

        for file in files2:
            rse_locks = [lock['rse_id'] for lock in get_replica_locks(scope=file['scope'], name=file['name'])]
            assert (rse_locks[0] == rse_locks[1])

    def test_add_rule_container_none(self, mock_scope, did_factory, jdoe_account):
        """ REPLICATION RULE (CORE): Add a replication rule on a container, NONE Grouping"""
        container = did_factory.random_container_did()
        add_did(did_type=DIDType.CONTAINER, account=jdoe_account, **container)
        all_files = []
        for i in range(3):
            files = create_files(3, mock_scope, self.rse1_id)
            all_files.extend(files)
            dataset = did_factory.random_dataset_did()
            add_did(did_type=DIDType.DATASET, account=jdoe_account, **dataset)
            attach_dids(dids=files, account=jdoe_account, **dataset)
            attach_dids(dids=[dataset], account=jdoe_account, **container)

        add_rule(dids=[container], account=jdoe_account, copies=1, rse_expression=self.T2, grouping='NONE', weight=None, lifetime=None, locked=False, subscription_id=None)
        for file in all_files:
            rse_locks = {lock['rse_id'] for lock in get_replica_locks(scope=file['scope'], name=file['name'])}
            assert self.rse3_id in rse_locks
            assert self.rse4_id not in rse_locks

    def test_add_rule_dataset_all(self, mock_scope, did_factory, jdoe_account):
        """ REPLICATION RULE (CORE): Add a replication rule on a dataset, ALL Grouping"""
        files = create_files(3, mock_scope, self.rse1_id)
        dataset = did_factory.random_dataset_did()
        add_did(did_type=DIDType.DATASET, account=jdoe_account, **dataset)
        attach_dids(dids=files, account=jdoe_account, **dataset)

        add_rule(dids=[dataset], account=jdoe_account, copies=2, rse_expression=self.T1, grouping='ALL', weight=None, lifetime=None, locked=False, subscription_id=None)

        # Check if the Locks are created properly
        t1 = {self.rse1_id, self.rse2_id, self.rse4_id}
        first_locks = None
        for file in files:
            if first_locks is None:
                first_locks = {lock['rse_id'] for lock in get_replica_locks(scope=file['scope'], name=file['name'])}
            rse_locks = {lock['rse_id'] for lock in get_replica_locks(scope=file['scope'], name=file['name'])}
            assert (len(t1.intersection(rse_locks)) == 2)
            assert (len(first_locks.intersection(rse_locks)) == 2)

        # Check if the DatasetLocks are created properly
        dataset_locks = [lock for lock in get_dataset_locks(**dataset)]
        assert (len(t1.intersection(set([lock['rse_id'] for lock in dataset_locks]))) == 2)
        assert (len(first_locks.intersection(set([lock['rse_id'] for lock in dataset_locks]))) == 2)

    def test_add_rule_container_all(self, mock_scope, did_factory, jdoe_account):
        """ REPLICATION RULE (CORE): Add a replication rule on a container, ALL Grouping"""
        container = did_factory.random_container_did()
        add_did(did_type=DIDType.CONTAINER, account=jdoe_account, **container)
        all_files = []
        for i in range(3):
            files = create_files(3, mock_scope, self.rse1_id)
            all_files.extend(files)
            dataset = did_factory.random_dataset_did()
            add_did(did_type=DIDType.DATASET, account=jdoe_account, **dataset)
            attach_dids(dids=files, account=jdoe_account, **dataset)
            attach_dids(dids=[dataset], account=jdoe_account, **container)

        add_rule(dids=[container], account=jdoe_account, copies=2, rse_expression=self.T1, grouping='ALL', weight=None, lifetime=None, locked=False, subscription_id=None)

        t1 = {self.rse1_id, self.rse2_id, self.rse4_id}
        first_locks = None
        for file in all_files:
            if first_locks is None:
                first_locks = {lock['rse_id'] for lock in get_replica_locks(scope=file['scope'], name=file['name'])}
            rse_locks = {lock['rse_id'] for lock in get_replica_locks(scope=file['scope'], name=file['name'])}
            assert (len(t1.intersection(rse_locks)) == 2)
            assert (len(first_locks.intersection(rse_locks)) == 2)

    def test_add_rule_requests(self, mock_scope, did_factory, jdoe_account):
        """ REPLICATION RULE (CORE): Add a replication rule on a dataset, DATASET Grouping"""
        files = create_files(3, mock_scope, self.rse1_id)
        dataset = did_factory.random_dataset_did()
        add_did(did_type=DIDType.DATASET, account=jdoe_account, **dataset)
        attach_dids(dids=files, account=jdoe_account, **dataset)

        add_rule(dids=[dataset], account=jdoe_account, copies=2, rse_expression=self.T1, grouping='DATASET', weight=None, lifetime=None, locked=False, subscription_id=None)

        # Check if the Locks are created properly
        t1 = {self.rse1_id, self.rse2_id, self.rse4_id}
        first_locks = None
        for file in files:
            if first_locks is None:
                first_locks = {lock['rse_id'] for lock in get_replica_locks(scope=file['scope'], name=file['name'])}
            rse_locks = {lock['rse_id'] for lock in get_replica_locks(scope=file['scope'], name=file['name'])}
            assert (len(t1.intersection(rse_locks)) == 2)
            assert (len(first_locks.intersection(rse_locks)) == 2)

        # Check if the DatasetLocks are created properly
        dataset_locks = [lock for lock in get_dataset_locks(**dataset)]
        assert (len(t1.intersection(set([lock['rse_id'] for lock in dataset_locks]))) == 2)
        assert (len(first_locks.intersection(set([lock['rse_id'] for lock in dataset_locks]))) == 2)

    def test_add_rule_dataset_dataset(self, mock_scope, did_factory, jdoe_account):
        """ REPLICATION RULE (CORE): Add a replication rule on a dataset and check if requests are created"""
        files = create_files(3, mock_scope, self.rse1_id)
        dataset = did_factory.random_dataset_did()
        add_did(did_type=DIDType.DATASET, account=jdoe_account, **dataset)
        attach_dids(dids=files, account=jdoe_account, **dataset)

        add_rule(dids=[dataset], account=jdoe_account, copies=1, rse_expression=self.rse4, grouping='DATASET', weight=None, lifetime=None, locked=False, subscription_id=None)

        for file in files:
            get_request_by_did(scope=file['scope'], name=file['name'], rse_id=self.rse4_id)

    def test_add_rule_container_dataset(self, mock_scope, did_factory, jdoe_account):
        """ REPLICATION RULE (CORE): Add a replication rule on a container, DATASET Grouping"""
        container = did_factory.random_container_did()
        add_did(did_type=DIDType.CONTAINER, account=jdoe_account, **container)
        all_files = []
        dataset_files = []
        for i in range(3):
            files = create_files(3, mock_scope, self.rse1_id)
            all_files.extend(files)
            dataset = did_factory.random_dataset_did()
            add_did(did_type=DIDType.DATASET, account=jdoe_account, **dataset)
            attach_dids(dids=files, account=jdoe_account, **dataset)
            attach_dids(dids=[dataset], account=jdoe_account, **container)
            dataset_files.append((dataset, files))

        add_rule(dids=[container], account=jdoe_account, copies=2, rse_expression=self.T1, grouping='DATASET', weight=None, lifetime=None, locked=False, subscription_id=None)

        t1 = {self.rse1_id, self.rse2_id, self.rse4_id}
        for ds, ds_files in dataset_files:
            first_locks = None
            for file in ds_files:
                if first_locks is None:
                    first_locks = {lock['rse_id'] for lock in get_replica_locks(scope=file['scope'], name=file['name'])}
                rse_locks = {lock['rse_id'] for lock in get_replica_locks(scope=file['scope'], name=file['name'])}
                assert (len(t1.intersection(rse_locks)) == 2)
                assert (len(first_locks.intersection(rse_locks)) == 2)

    def test_add_rule_dataset_none_with_weights(self, mock_scope, did_factory, jdoe_account):
        """ REPLICATION RULE (CORE): Add a replication rule on a dataset, NONE Grouping, WEIGHTS"""
        files = create_files(3, mock_scope, self.rse1_id)
        dataset = did_factory.random_dataset_did()
        add_did(did_type=DIDType.DATASET, account=jdoe_account, **dataset)
        attach_dids(dids=files, account=jdoe_account, **dataset)

        add_rule(dids=[dataset], account=jdoe_account, copies=2, rse_expression=self.T1, grouping='NONE', weight="fakeweight", lifetime=None, locked=False, subscription_id=None)

        # Check if the Locks are created properly
        t1 = {self.rse1_id, self.rse2_id, self.rse4_id}
        for file in files:
            rse_locks = {lock['rse_id'] for lock in get_replica_locks(scope=file['scope'], name=file['name'])}
            assert (len(t1.intersection(rse_locks)) == 2)
            assert self.rse1_id in rse_locks

    def test_add_rule_container_dataset_with_weights(self, mock_scope, did_factory, jdoe_account):
        """ REPLICATION RULE (CORE): Add a replication rule on a container, DATASET Grouping, WEIGHTS"""
        container = did_factory.random_container_did()
        add_did(did_type=DIDType.CONTAINER, account=jdoe_account, **container)
        all_files = []
        dataset_files = []
        for i in range(3):
            files = create_files(3, mock_scope, self.rse1_id)
            all_files.extend(files)
            dataset = did_factory.random_dataset_did()
            add_did(did_type=DIDType.DATASET, account=jdoe_account, **dataset)
            attach_dids(dids=files, account=jdoe_account, **dataset)
            attach_dids(dids=[dataset], account=jdoe_account, **container)
            dataset_files.append((dataset, files))

        add_rule(dids=[container], account=jdoe_account, copies=2, rse_expression=self.T1, grouping='DATASET', weight='fakeweight', lifetime=None, locked=False, subscription_id=None)

        t1 = {self.rse1_id, self.rse2_id, self.rse4_id}
        for ds, ds_files in dataset_files:
            first_locks = None
            for file in ds_files:
                if first_locks is None:
                    first_locks = {lock['rse_id'] for lock in get_replica_locks(scope=file['scope'], name=file['name'])}
                rse_locks = {lock['rse_id'] for lock in get_replica_locks(scope=file['scope'], name=file['name'])}
                assert (len(t1.intersection(rse_locks)) == 2)
                assert (len(first_locks.intersection(rse_locks)) == 2)
                assert self.rse1_id in rse_locks

    def test_get_rule(self, mock_scope, did_factory, jdoe_account):
        """ REPLICATION RULE (CORE): Test to get a previously created rule"""
        files = create_files(3, mock_scope, self.rse1_id)
        dataset = did_factory.random_dataset_did()
        add_did(did_type=DIDType.DATASET, account=jdoe_account, **dataset)
        attach_dids(dids=files, account=jdoe_account, **dataset)

        rule_id = add_rule(dids=[dataset], account=jdoe_account, copies=2, rse_expression=self.T1, grouping='NONE', weight='fakeweight', lifetime=None, locked=False, subscription_id=None)[0]
        assert (rule_id == get_rule(rule_id)['id'].replace('-', '').lower())
        pytest.raises(RuleNotFound, get_rule, uuid())

    def test_delete_rule(self, mock_scope, did_factory, jdoe_account):
        """ REPLICATION RULE (CORE): Test to delete a previously created rule"""
        files = create_files(3, mock_scope, self.rse1_id)
        dataset = did_factory.random_dataset_did()
        add_did(did_type=DIDType.DATASET, account=jdoe_account, **dataset)
        attach_dids(dids=files, account=jdoe_account, **dataset)

        rule_id = add_rule(dids=[dataset], account=jdoe_account, copies=2, rse_expression=self.T1, grouping='DATASET', weight='fakeweight', lifetime=None, locked=False, subscription_id=None)[0]
        delete_rule(rule_id)
        for file in files:
            rse_locks = get_replica_locks(scope=file['scope'], name=file['name'])
            assert (len(rse_locks) == 0)
        pytest.raises(RuleNotFound, delete_rule, uuid())

    def test_delete_rule_and_cancel_transfers(self, mock_scope, did_factory, jdoe_account):
        """ REPLICATION RULE (CORE): Test to delete a previously created rule and do not cancel overlapping transfers"""
        files = create_files(3, mock_scope, self.rse1_id)
        dataset = did_factory.random_dataset_did()
        add_did(did_type=DIDType.DATASET, account=jdoe_account, **dataset)
        attach_dids(dids=files, account=jdoe_account, **dataset)

        rule_id_1 = add_rule(dids=[dataset], account=jdoe_account, copies=1, rse_expression=self.rse1, grouping='NONE', weight='fakeweight', lifetime=None, locked=False, subscription_id=None)[0]
        add_rule(dids=[dataset], account=jdoe_account, copies=2, rse_expression=self.T1, grouping='NONE', weight='fakeweight', lifetime=None, locked=False, subscription_id=None)[0]
        add_rule(dids=[dataset], account=jdoe_account, copies=3, rse_expression=self.T1, grouping='NONE', weight='fakeweight', lifetime=None, locked=False, subscription_id=None)[0]

        delete_rule(rule_id_1)

        for file in files:
            rse_locks = get_replica_locks(scope=file['scope'], name=file['name'])
            assert (len(rse_locks) == 5)
            # TODO Need to check transfer queue here, this is actually not the check of this test case
        pytest.raises(RuleNotFound, delete_rule, uuid())

    def test_locked_rule(self, mock_scope, did_factory, jdoe_account):
        """ REPLICATION RULE (CORE): Delete a locked replication rule"""
        files = create_files(3, mock_scope, self.rse1_id)
        dataset = did_factory.random_dataset_did()
        add_did(did_type=DIDType.DATASET, account=jdoe_account, **dataset)
        attach_dids(dids=files, account=jdoe_account, **dataset)

        rule_id_1 = add_rule(dids=[dataset], account=jdoe_account, copies=1, rse_expression=self.rse1, grouping='NONE', weight='fakeweight', lifetime=None, locked=True, subscription_id=None)[0]

        pytest.raises(UnsupportedOperation, delete_rule, rule_id_1)
        update_rule(rule_id=rule_id_1, options={'locked': False})
        delete_rule(rule_id=rule_id_1)

    @pytest.mark.noparallel(reason='runs abacus account update')
    def test_account_counter_rule_create(self, mock_scope, did_factory, jdoe_account):
        """ REPLICATION RULE (CORE): Test if the account counter is updated correctly when new rule is created"""

        account_update(once=True)
        account_counter_before = get_usage(self.rse1_id, jdoe_account)

        files = create_files(3, mock_scope, self.rse1_id, bytes_=100)
        dataset = did_factory.random_dataset_did()
        add_did(did_type=DIDType.DATASET, account=jdoe_account, **dataset)
        attach_dids(dids=files, account=jdoe_account, **dataset)

        add_rule(dids=[dataset], account=jdoe_account, copies=1, rse_expression=self.rse1, grouping='ALL', weight=None, lifetime=None, locked=False, subscription_id=None)

        # Check if the counter has been updated correctly
        account_update(once=True)
        account_counter_after = get_usage(self.rse1_id, jdoe_account)
        assert (account_counter_before['bytes'] + 3 * 100 == account_counter_after['bytes'])
        assert (account_counter_before['files'] + 3 == account_counter_after['files'])

    @pytest.mark.noparallel(reason='runs abacus account update')
    def test_account_counter_rule_delete(self, mock_scope, did_factory, jdoe_account):
        """ REPLICATION RULE (CORE): Test if the account counter is updated correctly when a rule is removed"""

        files = create_files(3, mock_scope, self.rse1_id, bytes_=100)
        dataset = did_factory.random_dataset_did()
        add_did(did_type=DIDType.DATASET, account=jdoe_account, **dataset)
        attach_dids(dids=files, account=jdoe_account, **dataset)

        rule_id = add_rule(dids=[dataset], account=jdoe_account, copies=1, rse_expression=self.rse1, grouping='ALL', weight=None, lifetime=None, locked=False, subscription_id=None)[0]

        account_update(once=True)
        account_counter_before = get_usage(self.rse1_id, jdoe_account)

        delete_rule(rule_id)
        account_update(once=True)

        # Check if the counter has been updated correctly
        account_counter_after = get_usage(self.rse1_id, jdoe_account)
        assert (account_counter_before['bytes'] - 3 * 100 == account_counter_after['bytes'])
        assert (account_counter_before['files'] - 3 == account_counter_after['files'])

    @pytest.mark.noparallel(reason='runs abacus account update')
    def test_account_counter_rule_update(self, vo, mock_scope, did_factory, jdoe_account, root_account):
        """ REPLICATION RULE (CORE): Test if the account counter is updated correctly when a rule is updated"""

        files = create_files(3, mock_scope, self.rse1_id, bytes_=100)
        dataset = did_factory.random_dataset_did()
        add_did(did_type=DIDType.DATASET, account=jdoe_account, **dataset)
        attach_dids(dids=files, account=jdoe_account, **dataset)

        rule_id = add_rule(dids=[dataset], account=jdoe_account, copies=1, rse_expression=self.rse1, grouping='ALL', weight=None, lifetime=None, locked=False, subscription_id=None)[0]

        account_update(once=True)
        account_counter_before_1 = get_usage(self.rse1_id, jdoe_account)
        account_counter_before_2 = get_usage(self.rse1_id, root_account)

        rucio.gateway.rule.update_replication_rule(rule_id, {'account': 'root'}, issuer='root', vo=vo)
        account_update(once=True)

        # Check if the counter has been updated correctly
        account_counter_after_1 = get_usage(self.rse1_id, jdoe_account)
        account_counter_after_2 = get_usage(self.rse1_id, root_account)
        assert (account_counter_before_1['bytes'] - 3 * 100 == account_counter_after_1['bytes'])
        assert (account_counter_before_2['bytes'] + 3 * 100 == account_counter_after_2['bytes'])

    @pytest.mark.noparallel(reason='runs abacus rse update')
    def test_rse_counter_unavailable_replicas(self, mock_scope, did_factory, jdoe_account):
        """ REPLICATION RULE (CORE): Test if creating UNAVAILABLE replicas updates the RSE Counter correctly"""

        rse_update(once=True)
        rse_counter_before = get_rse_counter(self.rse2_id)

        files = create_files(3, mock_scope, self.rse1_id, bytes_=100)
        dataset = did_factory.random_dataset_did()
        add_did(did_type=DIDType.DATASET, account=jdoe_account, **dataset)
        attach_dids(dids=files, account=jdoe_account, **dataset)

        add_rule(dids=[dataset], account=jdoe_account, copies=1, rse_expression=self.rse2, grouping='ALL', weight=None, lifetime=None, locked=False, subscription_id=None)

        # Check if the rse has been updated correctly
        rse_update(once=True)
        rse_counter_after = get_rse_counter(self.rse2_id)
        assert (rse_counter_before['bytes'] + 3 * 100 == rse_counter_after['bytes'])
        assert (rse_counter_before['files'] + 3 == rse_counter_after['files'])

    def test_rule_add_fails_account_local_limit(self, mock_scope, did_factory, jdoe_account):
        """ REPLICATION RULE (CORE): Test if a rule fails correctly when local account limit conflict"""

        files = create_files(3, mock_scope, self.rse2_id, bytes_=100)
        # local quota
        dataset = did_factory.random_dataset_did()
        add_did(did_type=DIDType.DATASET, account=jdoe_account, **dataset)
        attach_dids(dids=files, account=jdoe_account, **dataset)

        set_local_account_limit(account=jdoe_account, rse_id=self.rse2_id, bytes_=5)

        pytest.raises(InsufficientAccountLimit, add_rule, dids=[dataset], account=jdoe_account, copies=1, rse_expression=self.rse2, grouping='ALL', weight=None, lifetime=None, locked=False, subscription_id=None)
        set_local_account_limit(account=jdoe_account, rse_id=self.rse2_id, bytes_=-1)

    def test_rule_add_fails_account_global_limit(self, mock_scope, did_factory, random_account, rse_factory):
        """ REPLICATION RULE (CORE): Test if a rule fails correctly when global account limit conflict"""
        rse1, rse1_id = rse_factory.make_mock_rse()
        rse2, rse2_id = rse_factory.make_mock_rse()
        rse3, rse3_id = rse_factory.make_mock_rse()
        files = create_files(3, mock_scope, rse1_id, bytes_=100)
        dataset = did_factory.random_dataset_did()
        add_did(did_type=DIDType.DATASET, account=random_account, **dataset)
        attach_dids(dids=files, account=random_account, **dataset)

        set_local_account_limit(account=random_account, rse_id=rse1_id, bytes_=400)
        # check with two global limits - one breaking limit is enough to let the rule fail
        set_global_account_limit(rse_expression=f'{rse1}|{rse2}', account=random_account, bytes_=400)
        set_global_account_limit(rse_expression=f'{rse1}|{rse3}', account=random_account, bytes_=10)
        pytest.raises(InsufficientAccountLimit, add_rule, dids=[dataset], account=random_account, copies=1, rse_expression=rse1, grouping='ALL', weight=None, lifetime=None, locked=False, subscription_id=None)

    def test_rule_add_fails_rse_limit(self, mock_scope, did_factory, jdoe_account):
        """ REPLICATION RULE (CORE): Test if a rule fails correctly when rse limit set"""

        files = create_files(3, mock_scope, self.rse1_id, bytes_=100)
        dataset = did_factory.random_dataset_did()
        add_did(did_type=DIDType.DATASET, account=jdoe_account, **dataset)
        attach_dids(dids=files, account=jdoe_account, **dataset)

        set_rse_limits(self.rse2_id, 'MaxSpaceAvailable', 250)
        try:
            pytest.raises(RSEOverQuota, add_rule, dids=[dataset], account=jdoe_account, copies=1, rse_expression=self.rse2, grouping='ALL', weight=None, lifetime=None, locked=False, subscription_id=None)
            pytest.raises(RSEOverQuota, add_rule, dids=[dataset], account=jdoe_account, copies=1, rse_expression=self.rse2, grouping='DATASET', weight=None, lifetime=None, locked=False, subscription_id=None)
            # TODO: fix #5845 and uncomment this line
            # pytest.raises(RSEOverQuota, add_rule, dids=[dataset], account=jdoe_account, copies=1, rse_expression=self.rse2, grouping='NONE', weight=None, lifetime=None, locked=False, subscription_id=None)
        finally:
            set_rse_limits(self.rse2_id, 'MaxSpaceAvailable', -1)

    def test_dataset_callback(self, mock_scope, did_factory, jdoe_account):
        """ REPLICATION RULE (CORE): Test dataset callback"""

        files = create_files(3, mock_scope, self.rse1_id, bytes_=100)
        dataset = did_factory.random_dataset_did()
        add_did(did_type=DIDType.DATASET, account=jdoe_account, **dataset)
        attach_dids(dids=files, account=jdoe_account, **dataset)

        set_status(open=False, **dataset)

        rule_id = add_rule(dids=[dataset], account=jdoe_account, copies=1, rse_expression=self.rse2, grouping='DATASET', weight=None, lifetime=None, locked=False, subscription_id=None, notify='C')[0]

        successful_transfer(scope=mock_scope, name=files[0]['name'], rse_id=self.rse2_id, nowait=False)
        successful_transfer(scope=mock_scope, name=files[1]['name'], rse_id=self.rse2_id, nowait=False)
        successful_transfer(scope=mock_scope, name=files[2]['name'], rse_id=self.rse2_id, nowait=False)

        # Check if rule exists
        assert (True is check_dataset_ok_callback(dataset['scope'], dataset['name'], self.rse2, self.rse2_id, rule_id))

    def test_dataset_callback_no(self, mock_scope, did_factory, jdoe_account):
        """ REPLICATION RULE (CORE): Test dataset callback should not be sent"""

        files = create_files(3, mock_scope, self.rse1_id, bytes_=100)
        dataset = did_factory.random_dataset_did()
        add_did(did_type=DIDType.DATASET, account=jdoe_account, **dataset)
        attach_dids(dids=files, account=jdoe_account, **dataset)

        set_status(open=False, **dataset)

        rule_id = add_rule(dids=[dataset], account=jdoe_account, copies=1, rse_expression=self.rse2, grouping='DATASET', weight=None, lifetime=None, locked=False, subscription_id=None, notify='C')[0]

        successful_transfer(scope=mock_scope, name=files[0]['name'], rse_id=self.rse2_id, nowait=False)
        successful_transfer(scope=mock_scope, name=files[1]['name'], rse_id=self.rse2_id, nowait=False)

        # Check if rule exists
        assert (False is check_dataset_ok_callback(dataset['scope'], dataset['name'], self.rse2, self.rse2_id, rule_id))

    def test_dataset_callback_close_late(self, mock_scope, did_factory, jdoe_account):
        """ REPLICATION RULE (CORE): Test dataset callback with late close"""

        files = create_files(3, mock_scope, self.rse1_id, bytes_=100)
        dataset = did_factory.random_dataset_did()
        add_did(did_type=DIDType.DATASET, account=jdoe_account, **dataset)
        attach_dids(dids=files, account=jdoe_account, **dataset)

        rule_id = add_rule(dids=[dataset], account=jdoe_account, copies=1, rse_expression=self.rse2, grouping='DATASET', weight=None, lifetime=None, locked=False, subscription_id=None, notify='C')[0]

        successful_transfer(scope=mock_scope, name=files[0]['name'], rse_id=self.rse2_id, nowait=False)
        successful_transfer(scope=mock_scope, name=files[1]['name'], rse_id=self.rse2_id, nowait=False)
        successful_transfer(scope=mock_scope, name=files[2]['name'], rse_id=self.rse2_id, nowait=False)

        # Check if rule exists
        assert (False is check_dataset_ok_callback(dataset['scope'], dataset['name'], self.rse2, self.rse2_id, rule_id))
        set_status(open=False, **dataset)
        assert (True is check_dataset_ok_callback(dataset['scope'], dataset['name'], self.rse2, self.rse2_id, rule_id))

    @pytest.mark.noparallel(reason='runs re_evaluator')
    def test_dataset_callback_with_evaluator(self, mock_scope, did_factory, jdoe_account):
        """ REPLICATION RULE (CORE): Test dataset callback with judge evaluator"""

        files = create_files(3, mock_scope, self.rse1_id, bytes_=100)
        dataset = did_factory.random_dataset_did()
        add_did(did_type=DIDType.DATASET, account=jdoe_account, **dataset)

        rule_id = add_rule(dids=[dataset], account=jdoe_account, copies=1, rse_expression=self.rse2, grouping='DATASET', weight=None, lifetime=None, locked=False, subscription_id=None, notify='C')[0]

        assert (False is check_dataset_ok_callback(dataset['scope'], dataset['name'], self.rse2, self.rse2_id, rule_id))

        attach_dids(dids=files, account=jdoe_account, **dataset)
        set_status(open=False, **dataset)
        assert (False is check_dataset_ok_callback(dataset['scope'], dataset['name'], self.rse2, self.rse2_id, rule_id))

        re_evaluator(once=True, did_limit=None)

        successful_transfer(scope=mock_scope, name=files[0]['name'], rse_id=self.rse2_id, nowait=False)
        successful_transfer(scope=mock_scope, name=files[1]['name'], rse_id=self.rse2_id, nowait=False)
        successful_transfer(scope=mock_scope, name=files[2]['name'], rse_id=self.rse2_id, nowait=False)

        assert (True is check_dataset_ok_callback(dataset['scope'], dataset['name'], self.rse2, self.rse2_id, rule_id))

    @pytest.mark.noparallel(reason='runs re_evaluator')
    def test_rule_progress_callback_with_evaluator(self, mock_scope, did_factory, jdoe_account):
        """ REPLICATION RULE (CORE): Test rule progress callback with judge evaluator"""

        files = create_files(30, mock_scope, self.rse1_id, bytes_=100)
        dataset = did_factory.random_dataset_did()
        add_did(did_type=DIDType.DATASET, account=jdoe_account, **dataset)

        rule_id = add_rule(dids=[dataset], account=jdoe_account, copies=1, rse_expression=self.rse2, grouping='DATASET', weight=None, lifetime=None, locked=False, subscription_id=None, notify='P')[0]

        assert (False is check_rule_progress_callback(dataset['scope'], dataset['name'], 0, rule_id))

        attach_dids(dids=files, account=jdoe_account, **dataset)
        re_evaluator(once=True, did_limit=None)

        set_status(open=False, **dataset)
        assert (False is check_rule_progress_callback(dataset['scope'], dataset['name'], 0, rule_id))

        successful_transfer(scope=mock_scope, name=files[0]['name'], rse_id=self.rse2_id, nowait=False)
        assert (False is check_rule_progress_callback(dataset['scope'], dataset['name'], 10, rule_id))
        successful_transfer(scope=mock_scope, name=files[1]['name'], rse_id=self.rse2_id, nowait=False)
        assert (False is check_rule_progress_callback(dataset['scope'], dataset['name'], 10, rule_id))
        successful_transfer(scope=mock_scope, name=files[2]['name'], rse_id=self.rse2_id, nowait=False)
        assert (True is check_rule_progress_callback(dataset['scope'], dataset['name'], 10, rule_id))
        successful_transfer(scope=mock_scope, name=files[3]['name'], rse_id=self.rse2_id, nowait=False)
        assert (False is check_rule_progress_callback(dataset['scope'], dataset['name'], 20, rule_id))
        successful_transfer(scope=mock_scope, name=files[4]['name'], rse_id=self.rse2_id, nowait=False)
        successful_transfer(scope=mock_scope, name=files[5]['name'], rse_id=self.rse2_id, nowait=False)
        successful_transfer(scope=mock_scope, name=files[6]['name'], rse_id=self.rse2_id, nowait=False)
        successful_transfer(scope=mock_scope, name=files[7]['name'], rse_id=self.rse2_id, nowait=False)
        successful_transfer(scope=mock_scope, name=files[8]['name'], rse_id=self.rse2_id, nowait=False)
        assert (True is check_rule_progress_callback(dataset['scope'], dataset['name'], 30, rule_id))
        successful_transfer(scope=mock_scope, name=files[9]['name'], rse_id=self.rse2_id, nowait=False)
        successful_transfer(scope=mock_scope, name=files[10]['name'], rse_id=self.rse2_id, nowait=False)
        successful_transfer(scope=mock_scope, name=files[11]['name'], rse_id=self.rse2_id, nowait=False)
        successful_transfer(scope=mock_scope, name=files[12]['name'], rse_id=self.rse2_id, nowait=False)
        successful_transfer(scope=mock_scope, name=files[13]['name'], rse_id=self.rse2_id, nowait=False)
        successful_transfer(scope=mock_scope, name=files[14]['name'], rse_id=self.rse2_id, nowait=False)
        successful_transfer(scope=mock_scope, name=files[15]['name'], rse_id=self.rse2_id, nowait=False)
        successful_transfer(scope=mock_scope, name=files[16]['name'], rse_id=self.rse2_id, nowait=False)
        successful_transfer(scope=mock_scope, name=files[17]['name'], rse_id=self.rse2_id, nowait=False)
        assert (True is check_rule_progress_callback(dataset['scope'], dataset['name'], 60, rule_id))
        successful_transfer(scope=mock_scope, name=files[18]['name'], rse_id=self.rse2_id, nowait=False)
        successful_transfer(scope=mock_scope, name=files[19]['name'], rse_id=self.rse2_id, nowait=False)
        successful_transfer(scope=mock_scope, name=files[20]['name'], rse_id=self.rse2_id, nowait=False)
        successful_transfer(scope=mock_scope, name=files[21]['name'], rse_id=self.rse2_id, nowait=False)
        successful_transfer(scope=mock_scope, name=files[22]['name'], rse_id=self.rse2_id, nowait=False)
        successful_transfer(scope=mock_scope, name=files[23]['name'], rse_id=self.rse2_id, nowait=False)
        successful_transfer(scope=mock_scope, name=files[24]['name'], rse_id=self.rse2_id, nowait=False)
        successful_transfer(scope=mock_scope, name=files[25]['name'], rse_id=self.rse2_id, nowait=False)
        successful_transfer(scope=mock_scope, name=files[26]['name'], rse_id=self.rse2_id, nowait=False)
        assert (True is check_rule_progress_callback(dataset['scope'], dataset['name'], 90, rule_id))
        successful_transfer(scope=mock_scope, name=files[27]['name'], rse_id=self.rse2_id, nowait=False)
        successful_transfer(scope=mock_scope, name=files[28]['name'], rse_id=self.rse2_id, nowait=False)
        successful_transfer(scope=mock_scope, name=files[29]['name'], rse_id=self.rse2_id, nowait=False)
        assert (True is check_rule_progress_callback(dataset['scope'], dataset['name'], 100, rule_id))

    def test_add_rule_with_purge(self, mock_scope, did_factory, jdoe_account):
        """ REPLICATION RULE (CORE): Add a replication rule with purge setting"""
        files = create_files(3, mock_scope, self.rse1_id)
        dataset = did_factory.random_dataset_did()
        add_did(did_type=DIDType.DATASET, account=jdoe_account, **dataset)
        attach_dids(dids=files, account=jdoe_account, **dataset)

        rule_id = add_rule(dids=[dataset], account=jdoe_account, copies=1, rse_expression=self.rse3, grouping='NONE', weight=None, lifetime=None, locked=False, subscription_id=None, purge_replicas=True)[0]

        delete_rule(rule_id)

        # Check if the Locks are created properly
        for file in files:
            replica = get_replica(rse_id=self.rse3_id, scope=file['scope'], name=file['name'])
            assert (replica['tombstone'] == OBSOLETE)

    def test_add_rule_with_ignore_availability(self, vo, mock_scope, did_factory, jdoe_account):
        """ REPLICATION RULE (CORE): Add a replication rule with ignore_availability setting"""
        rse = rse_name_generator()
        rse_id = add_rse(rse, vo=vo)
        update_rse(rse_id, {'availability_write': False})
        set_local_account_limit(jdoe_account, rse_id, -1)

        files = create_files(3, mock_scope, self.rse1_id)
        dataset = did_factory.random_dataset_did()
        add_did(did_type=DIDType.DATASET, account=jdoe_account, **dataset)
        attach_dids(dids=files, account=jdoe_account, **dataset)

        with pytest.raises(RSEWriteBlocked):
            add_rule(dids=[dataset], account=jdoe_account, copies=1, rse_expression=rse, grouping='NONE', weight=None, lifetime=None, locked=False, subscription_id=None)[0]

        add_rule(dids=[dataset], account=jdoe_account, copies=1, rse_expression=rse, grouping='NONE', weight=None, lifetime=None, locked=False, subscription_id=None, ignore_availability=True)[0]
        for file in files:
            for filtered_lock in [lock for lock in get_replica_locks(scope=file['scope'], name=file['name'])]:
                assert (filtered_lock['state'] == LockState.STUCK)

    @pytest.mark.skipif(os.environ.get('POLICY') != 'atlas', reason='ATLAS-specific test')
    def test_delete_rule_country_admin(self, vo, mock_scope, did_factory, jdoe_account):
        """ REPLICATION RULE (CORE): Delete a rule with a country admin account"""
        rse = rse_name_generator()
        rse_id = add_rse(rse, vo=vo)
        add_rse_attribute(rse_id, RseAttr.COUNTRY, 'test')
        set_local_account_limit(jdoe_account, rse_id, -1)

        files = create_files(3, mock_scope, self.rse1_id)
        dataset = did_factory.random_dataset_did()
        add_did(did_type=DIDType.DATASET, account=jdoe_account, **dataset)
        attach_dids(dids=files, account=jdoe_account, **dataset)

        rule_id = add_rule(dids=[dataset], account=jdoe_account, copies=1, rse_expression=rse, grouping='NONE', weight=None, lifetime=None, locked=False, subscription_id=None)[0]

        usr = account_name_generator()
        add_account(usr, 'USER', 'rucio@email.com', 'root', vo=vo)

        with pytest.raises(AccessDenied):
            rucio.gateway.rule.delete_replication_rule(rule_id=rule_id, purge_replicas=None, issuer=usr, vo=vo)

        add_account_attribute(InternalAccount(usr, vo=vo), 'country-test', 'admin')
        rucio.gateway.rule.delete_replication_rule(rule_id=rule_id, purge_replicas=None, issuer=usr, vo=vo)

    def test_reduce_rule(self, mock_scope, did_factory, jdoe_account):
        """ REPLICATION RULE (CORE): Reduce a rule"""
        files = create_files(3, mock_scope, [self.rse1_id, self.rse2_id])
        dataset = did_factory.random_dataset_did()
        add_did(did_type=DIDType.DATASET, account=jdoe_account, **dataset)
        attach_dids(dids=files, account=jdoe_account, **dataset)

        rule_id = add_rule(dids=[dataset], account=jdoe_account, copies=2, rse_expression=self.rse1 + '|' + self.rse2, grouping='DATASET', weight=None, lifetime=None, locked=False, subscription_id=None)[0]

        assert (get_rule(rule_id)['state'] == RuleState.OK)

        rule_id2 = reduce_rule(rule_id=rule_id, copies=1, exclude_expression=self.rse1)

        assert (get_rule(rule_id2)['state'] == RuleState.OK)
        pytest.raises(RuleNotFound, get_rule, rule_id)

        files = create_files(3, mock_scope, [self.rse1_id, self.rse2_id])
        dataset = did_factory.random_dataset_did()
        add_did(did_type=DIDType.DATASET, account=jdoe_account, **dataset)
        attach_dids(dids=files, account=jdoe_account, **dataset)

        rule_id = add_rule(dids=[dataset], account=jdoe_account, copies=2, rse_expression=self.rse1 + '|' + self.rse2 + '|' + self.rse4, grouping='DATASET', weight=None, lifetime=None, locked=False, subscription_id=None)[0]

        with pytest.raises(RuleReplaceFailed):
            reduce_rule(rule_id=rule_id, copies=1, exclude_expression=self.rse1 + '|' + self.rse2)

    def test_move_rule(self, mock_scope, did_factory, jdoe_account):
        """ REPLICATION RULE (CORE): Move a rule"""
        files = create_files(3, mock_scope, [self.rse1_id])
        dataset = did_factory.random_dataset_did()
        add_did(did_type=DIDType.DATASET, account=jdoe_account, **dataset)
        attach_dids(dids=files, account=jdoe_account, **dataset)

        rule_id = add_rule(dids=[dataset], account=jdoe_account, copies=1, rse_expression=self.rse1, grouping='DATASET', weight=None, lifetime=None, locked=False, subscription_id=None)[0]

        assert (get_rule(rule_id)['state'] == RuleState.OK)

        rule_id2 = move_rule(rule_id, self.rse2)

        assert (get_rule(rule_id2)['state'] == RuleState.REPLICATING)
        assert (get_rule(rule_id)['child_rule_id'] == rule_id2)
        assert (get_rule(rule_id2)['activity'] == get_rule(rule_id)['activity'])
        assert (get_rule(rule_id2)['source_replica_expression'] == get_rule(rule_id)['source_replica_expression'])

        pytest.raises(RuleReplaceFailed, move_rule, rule_id, self.rse3)

    def test_move_rule_with_arguments(self, mock_scope, did_factory, jdoe_account):
        """ REPLICATION RULE (CORE): Move a rule with activity and source-replica-expression specified"""
        files = create_files(3, mock_scope, [self.rse1_id])
        dataset = did_factory.random_dataset_did()
        add_did(did_type=DIDType.DATASET, account=jdoe_account, **dataset)
        attach_dids(dids=files, account=jdoe_account, **dataset)

        rule_id = add_rule(dids=[dataset], account=jdoe_account, copies=1, rse_expression=self.rse1, grouping='DATASET', weight=None, lifetime=None, locked=False, subscription_id=None)[0]

        assert (get_rule(rule_id)['state'] == RuleState.OK)

        activity = "No User Subscriptions"
        source_replica_expression = self.rse2 + "|" + self.rse1
        rule_id2 = move_rule(rule_id, self.rse2, override={'activity': activity, 'source_replica_expression': source_replica_expression})

        assert (get_rule(rule_id2)['state'] == RuleState.REPLICATING)
        assert (get_rule(rule_id)['child_rule_id'] == rule_id2)
        assert (get_rule(rule_id2)['activity'] == activity)
        assert (get_rule(rule_id2)['source_replica_expression'] == source_replica_expression)

        pytest.raises(RuleReplaceFailed, move_rule, rule_id, self.rse3)
        pytest.raises(RucioException, move_rule, 'foo', self.rse3)

    @pytest.mark.skipif(os.environ.get('POLICY') != 'atlas', reason='ATLAS-specific test')
    def test_add_rule_with_scratchdisk(self, vo, mock_scope, did_factory, jdoe_account):
        """ REPLICATION RULE (CORE): Add a replication rule for scratchdisk"""
        rse = rse_name_generator()
        rse_id = add_rse(rse, vo=vo)
        add_rse_attribute(rse_id, RseAttr.TYPE, 'SCRATCHDISK')
        set_local_account_limit(jdoe_account, rse_id, -1)

        files = create_files(3, mock_scope, self.rse1_id)
        dataset = did_factory.random_dataset_did()
        add_did(did_type=DIDType.DATASET, account=jdoe_account, **dataset)
        attach_dids(dids=files, account=jdoe_account, **dataset)

        rule_id = add_rule(dids=[dataset], account=jdoe_account, copies=1, rse_expression='%s' % rse, grouping='DATASET', weight=None, lifetime=None, locked=False, subscription_id=None)[0]
        assert (get_rule(rule_id)['expires_at'] is not None)

        rule_id = add_rule(dids=[dataset], account=jdoe_account, copies=1, rse_expression='%s' % self.rse1, grouping='DATASET', weight=None, lifetime=None, locked=False, subscription_id=None)[0]
        assert (get_rule(rule_id)['expires_at'] is None)

    def test_add_rule_with_auto_approval(self, vo, mock_scope, did_factory, jdoe_account):
        """ REPLICATION RULE (CORE): Add a replication rule with auto approval"""
        rse = rse_name_generator()
        rse_id = add_rse(rse, vo=vo)

        files = create_files(3, mock_scope, self.rse1_id, bytes_=200)
        dataset = did_factory.random_dataset_did()
        add_did(did_type=DIDType.DATASET, account=jdoe_account, **dataset)
        attach_dids(dids=files, account=jdoe_account, **dataset)
        set_status(open=False, **dataset)

        with pytest.raises(InsufficientAccountLimit):
            rule_id = add_rule(dids=[dataset], account=jdoe_account, copies=1, rse_expression='%s' % rse, grouping='DATASET', weight=None, lifetime=None, locked=False, subscription_id=None)[0]

        rule_id = add_rule(dids=[dataset], account=jdoe_account, copies=1, rse_expression='%s' % rse, grouping='DATASET', weight=None, lifetime=None, locked=False, subscription_id=None, ask_approval=True)[0]
        assert (get_rule(rule_id)['state'] == RuleState.WAITING_APPROVAL)
        delete_rule(rule_id=rule_id)

        add_rse_attribute(rse_id, RseAttr.AUTO_APPROVE_BYTES, 500)
        rule_id = add_rule(dids=[dataset], account=jdoe_account, copies=1, rse_expression='%s' % rse, grouping='DATASET', weight=None, lifetime=None, locked=False, subscription_id=None, ask_approval=True)[0]
        assert (get_rule(rule_id)['state'] == RuleState.WAITING_APPROVAL)
        delete_rule(rule_id=rule_id)

        del_rse_attribute(rse_id, RseAttr.AUTO_APPROVE_BYTES)
        add_rse_attribute(rse_id, RseAttr.AUTO_APPROVE_BYTES, 1000)
        rule_id = add_rule(dids=[dataset], account=jdoe_account, copies=1, rse_expression='%s' % rse, grouping='DATASET', weight=None, lifetime=None, locked=False, subscription_id=None, ask_approval=True)[0]
        assert (get_rule(rule_id)['state'] == RuleState.INJECT)

    def test_add_rule_with_manual_approval_block(self, vo, mock_scope, did_factory, jdoe_account):
        """ REPLICATION RULE (CORE): Add a replication rule for a RSE with manual approval block"""
        rse = rse_name_generator()
        rse_id = add_rse(rse, vo=vo)
        add_rse_attribute(rse_id, RseAttr.BLOCK_MANUAL_APPROVAL, '1')
        set_local_account_limit(jdoe_account, rse_id, -1)

        files = create_files(3, mock_scope, self.rse1_id)
        dataset = did_factory.random_dataset_did()
        add_did(did_type=DIDType.DATASET, account=jdoe_account, **dataset)
        attach_dids(dids=files, account=jdoe_account, **dataset)

        with pytest.raises(ManualRuleApprovalBlocked):
            add_rule(dids=[dataset], account=jdoe_account, copies=1, rse_expression='%s' % rse, grouping='DATASET', weight=None, lifetime=None, locked=False, subscription_id=None, ask_approval=True)[0]

    def test_update_rule_child_rule(self, mock_scope, did_factory, jdoe_account):
        """ REPLICATION RULE (CORE): Update a replication rule with a child_rule"""
        files = create_files(3, mock_scope, self.rse1_id)
        dataset1 = did_factory.random_dataset_did()
        dataset2 = did_factory.random_dataset_did()
        add_did(did_type=DIDType.DATASET, account=jdoe_account, **dataset1)
        attach_dids(dids=files, account=jdoe_account, **dataset1)
        add_did(did_type=DIDType.DATASET, account=jdoe_account, **dataset2)
        attach_dids(dids=files, account=jdoe_account, **dataset2)

        rule_id_1 = add_rule(dids=[dataset1], account=jdoe_account, copies=1, rse_expression=self.rse1, grouping='DATASET', weight=None, lifetime=None, locked=False, subscription_id=None)[0]
        rule_id_2 = add_rule(dids=[dataset2], account=jdoe_account, copies=1, rse_expression=self.rse1, grouping='DATASET', weight=None, lifetime=None, locked=False, subscription_id=None)[0]
        rule_id_3 = add_rule(dids=[dataset1], account=jdoe_account, copies=1, rse_expression=self.rse2, grouping='DATASET', weight=None, lifetime=None, locked=False, subscription_id=None)[0]

        with pytest.raises(InputValidationError):
            update_rule(rule_id_1, options={'child_rule_id': rule_id_2})
        update_rule(rule_id_1, options={'child_rule_id': rule_id_3})
        with pytest.raises(UnsupportedOperation):
            delete_rule(rule_id_1)

    def test_update_rule_unset_child_rule(self, mock_scope, did_factory, jdoe_account):
        """ REPLICATION RULE (CORE): Update a replication rule with a child_rule"""
        files = create_files(3, mock_scope, self.rse1_id)
        dataset1 = did_factory.random_dataset_did()
        dataset2 = did_factory.random_dataset_did()
        add_did(did_type=DIDType.DATASET, account=jdoe_account, **dataset1)
        attach_dids(dids=files, account=jdoe_account, **dataset1)
        add_did(did_type=DIDType.DATASET, account=jdoe_account, **dataset2)
        attach_dids(dids=files, account=jdoe_account, **dataset2)

        parent_id: str = add_rule(
            rse_expression=self.rse1,
            dids=[dataset1], account=jdoe_account, copies=1,
            grouping='DATASET', weight=None, lifetime=None,
            locked=False, subscription_id=None
        )[0]

        # assign a child rule id
        # with move_rule: this sets an expiry date
        child_id = move_rule(parent_id, self.rse3)
        assert child_id
        assert get_rule(parent_id)['child_rule_id']
        assert get_rule(parent_id)['expires_at']

        # detach this child
        update_rule(parent_id, {'child_rule_id': None})
        assert not get_rule(parent_id)['child_rule_id']
        assert not get_rule(parent_id)['expires_at']

        # test detaching when there is no more attached child
        with pytest.raises(InputValidationError):
            update_rule(parent_id, {'child_rule_id': None})

    def test_rule_priority_set_and_update(self, mock_scope, did_factory, jdoe_account):
        files = create_files(1, mock_scope, self.rse1_id)
        dataset = did_factory.random_dataset_did()
        add_did(did_type=DIDType.DATASET, account=jdoe_account, **dataset)
        attach_dids(dids=files, account=jdoe_account, **dataset)

        rule_id = add_rule(dids=[dataset], account=jdoe_account, copies=1, rse_expression=self.rse2, grouping='NONE',
                           weight='fakeweight', lifetime=None, locked=False, subscription_id=None)[0]
        assert get_rule(rule_id)['priority'] == 3
        request = get_request_by_did(scope=files[0]['scope'], name=files[0]['name'], rse_id=self.rse2_id)
        assert request['priority'] == 3

        update_rule(rule_id, {'priority': 5})
        assert get_rule(rule_id)['priority'] == 5
        assert get_request_by_did(scope=files[0]['scope'], name=files[0]['name'], rse_id=self.rse2_id)['priority'] == 5

    def test_release_rule(self, mock_scope, did_factory, jdoe_account):
        """ REPLICATION RULE (CORE): Test to release a parent rule after child rule is OK"""
        files = create_files(3, mock_scope, self.rse1_id, bytes_=100)
        dataset = did_factory.random_dataset_did()
        add_did(did_type=DIDType.DATASET, account=jdoe_account, **dataset)
        attach_dids(dids=files, account=jdoe_account, **dataset)

        rule_id_1 = add_rule(dids=[dataset], account=jdoe_account, copies=1, rse_expression=self.rse1, grouping='DATASET', weight=None, lifetime=None, locked=False, subscription_id=None)[0]
        rule_id_2 = add_rule(dids=[dataset], account=jdoe_account, copies=1, rse_expression=self.rse2, grouping='DATASET', weight=None, lifetime=None, locked=False, subscription_id=None)[0]

        update_rule(rule_id_1, options={'child_rule_id': rule_id_2})

        with pytest.raises(UnsupportedOperation):
            delete_rule(rule_id_1)

        successful_transfer(scope=mock_scope, name=files[0]['name'], rse_id=self.rse2_id, nowait=False)
        with pytest.raises(UnsupportedOperation):
            delete_rule(rule_id_1)
        successful_transfer(scope=mock_scope, name=files[1]['name'], rse_id=self.rse2_id, nowait=False)
        with pytest.raises(UnsupportedOperation):
            delete_rule(rule_id_1)
        successful_transfer(scope=mock_scope, name=files[2]['name'], rse_id=self.rse2_id, nowait=False)
        delete_rule(rule_id_1)

    def test_metadata__rule(self, mock_scope, did_factory, jdoe_account):
        """ REPLICATION RULE (CORE): Test to write wfms metadata to rule"""
        files = create_files(3, mock_scope, self.rse1_id)
        dataset = did_factory.random_dataset_did()
        add_did(did_type=DIDType.DATASET, account=jdoe_account, **dataset)
        attach_dids(dids=files, account=jdoe_account, **dataset)

        rule_id = add_rule(dids=[dataset], account=jdoe_account, copies=2, rse_expression=self.T1, grouping='NONE',
                           weight='fakeweight', lifetime=None, locked=False, meta={'task_id': 55, 'job_ids': [1, 2, 3, 4]}, subscription_id=None)[0]
        assert (get_rule(rule_id)['meta'] == json.dumps({'task_id': 55, 'job_ids': [1, 2, 3, 4]}))

    def test_rule_on_archive(self, mock_scope, jdoe_account):
        """ REPLICATION RULE (CORE): Test to add a rule on a constituent should add rule on archive"""
        archive = {'scope': mock_scope, 'name': '%s.zip' % str(uuid()), 'type': 'FILE',
                   'bytes': 2596, 'adler32': 'beefdead'}
        add_replica(rse_id=self.rse1_id, scope=mock_scope, name=archive['name'], bytes_=2596, account=jdoe_account)
        files_in_archive = [{'scope': mock_scope, 'name': 'witrep-%i-%s' % (i, str(uuid())), 'type': 'FILE',
                             'bytes': 1234, 'adler32': 'deadbeef'} for i in range(2)]
        attach_dids(mock_scope, archive['name'], files_in_archive, jdoe_account)

        add_rule(dids=[{'scope': mock_scope, 'name': files_in_archive[1]['name']}], account=jdoe_account, copies=1, rse_expression=self.rse2, grouping='NONE',
                 weight=None, lifetime=None, locked=False, subscription_id=None)
        assert (len(list(list_rules(filters={'scope': mock_scope, 'name': archive['name']}))) == 1)

        # Check the same but now a replica of the constituent exists as well
        archive = {'scope': mock_scope, 'name': '%s.zip' % str(uuid()), 'type': 'FILE',
                   'bytes': 2596, 'adler32': 'beefdead'}
        add_replica(rse_id=self.rse1_id, scope=mock_scope, name=archive['name'], bytes_=2596, account=jdoe_account)
        files_in_archive = [{'scope': mock_scope, 'name': 'witrep-%i-%s' % (i, str(uuid())), 'type': 'FILE',
                             'bytes': 1234, 'adler32': 'deadbeef'} for i in range(2)]
        attach_dids(mock_scope, archive['name'], files_in_archive, jdoe_account)
        add_replica(rse_id=self.rse1_id, scope=mock_scope, name=files_in_archive[1]['name'], bytes_=2596, account=jdoe_account)

        add_rule(dids=[{'scope': mock_scope, 'name': files_in_archive[1]['name']}], account=jdoe_account, copies=1, rse_expression=self.rse2, grouping='NONE',
                 weight=None, lifetime=None, locked=False, subscription_id=None)
        assert (len(list(list_rules(filters={'scope': mock_scope, 'name': archive['name']}))) == 0)
        assert (len(list(list_rules(filters={'scope': mock_scope, 'name': files_in_archive[1]['name']}))) == 1)

    def test_add_rule_overlapping_dids(self, vo, mock_scope, jdoe_account):
        """ REPLICATION RULE (CORE): Test various overlap cases"""

        def mktree(scope, account):
            # container1213 = container12 + container13
            # container12 = ds1 + ds2
            # container13 = ds1 + ds3
            # ds1 = file1 .. file10
            # ds2 = file11 .. file20
            # ds3 = file1, file2, file11, file12, file21 .. file25
            # 11 replicas @ rse1  -> file1 ..file7,   file21 .. file24
            #  3 replicas @ rse2 -> file8 .. file10
            #  6 replicas @ rse3 -> file11 .. file16
            #  5 replicas @ rse4 -> file17 .. file20, file25
            for i in range(1, 8):
                add_replica(rse_id=get_rse_id(self.rse1, vo=vo), scope=scope, name='file_%06d.data' % i, bytes_=10000 + i, account=account)
            for i in range(8, 11):
                add_replica(rse_id=get_rse_id(self.rse2, vo=vo), scope=scope, name='file_%06d.data' % i, bytes_=10000 + i, account=account)
            for i in range(11, 17):
                add_replica(rse_id=get_rse_id(self.rse3, vo=vo), scope=scope, name='file_%06d.data' % i, bytes_=10000 + i, account=account)
            for i in range(17, 21):
                add_replica(rse_id=get_rse_id(self.rse4, vo=vo), scope=scope, name='file_%06d.data' % i, bytes_=10000 + i, account=account)
            for i in range(21, 25):
                add_replica(rse_id=get_rse_id(self.rse1, vo=vo), scope=scope, name='file_%06d.data' % i, bytes_=10000 + i, account=account)
            for i in range(25, 26):
                add_replica(rse_id=get_rse_id(self.rse4, vo=vo), scope=scope, name='file_%06d.data' % i, bytes_=10000 + i, account=account)

            add_did(scope=scope, name='ds1', did_type='DATASET', account=account)
            attach_dids(scope=scope, name='ds1', dids=[{'scope': scope, 'name': 'file_%06d.data' % i} for i in range(1, 10 + 1)], account=account)
            add_did(scope=scope, name='ds2', did_type='DATASET', account=account)
            attach_dids(scope=scope, name='ds2', dids=[{'scope': scope, 'name': 'file_%06d.data' % i} for i in range(11, 20 + 1)], account=account)
            add_did(scope=scope, name='ds3', did_type='DATASET', account=account)
            attach_dids(scope=scope, name='ds3', dids=[{'scope': scope, 'name': 'file_%06d.data' % i} for i in (list(range(21, 25 + 1)) + [1, 2, 11, 12])], account=account)
            add_did(scope=scope, name='container12', did_type='CONTAINER', account=account)
            attach_dids(scope=scope, name='container12', dids=[{'scope': scope, 'name': 'ds1'}, {'scope': scope, 'name': 'ds2'}, ], account=account)
            add_did(scope=scope, name='container13', did_type='CONTAINER', account=account)
            attach_dids(scope=scope, name='container13', dids=[{'scope': scope, 'name': 'ds1'}, {'scope': scope, 'name': 'ds3'}, ], account=account)
            add_did(scope=scope, name='container1213', did_type='CONTAINER', account=account)
            attach_dids(scope=scope, name='container1213', dids=[{'scope': scope, 'name': 'container12'}, {'scope': scope, 'name': 'container13'}, ], account=account)

        account = jdoe_account

        # test1 : ALL grouping -> select rse1 for all 3 datasets
        scope = InternalScope(('scope1_' + str(uuid()))[:21], vo=vo)  # scope field has max 25 chars including VO
        add_scope(scope, account)
        mktree(scope, account)
        rule_ids = add_rule(dids=[{'scope': scope, 'name': 'container1213'}], copies=1, rse_expression=f'{self.rse1}|{self.rse2}|{self.rse3}|{self.rse4}',
                            grouping='ALL', account=account, weight=None, lifetime=None, locked=False, subscription_id=None)
        rule = get_rule(rule_ids[0])
        print(rule['locks_ok_cnt'], rule['locks_replicating_cnt'])
        assert (rule['locks_ok_cnt'] == 11)
        assert (rule['locks_replicating_cnt'] == 14)
        dsl1 = list(get_dataset_locks(scope, 'ds1'))
        dsl2 = list(get_dataset_locks(scope, 'ds2'))
        dsl3 = list(get_dataset_locks(scope, 'ds3'))
        print(dsl1)
        print(dsl2)
        print(dsl3)
        assert (len(dsl1) == 1 and dsl1[0]['rse'] == self.rse1)
        assert (len(dsl2) == 1 and dsl2[0]['rse'] == self.rse1)
        assert (len(dsl3) == 1 and dsl3[0]['rse'] == self.rse1)

        # test2 : DATASET grouping -> select rse1 for ds1, rse3 for ds2 and rse1 for ds3
        scope = InternalScope(('scope2_' + str(uuid()))[:21], vo=vo)  # scope field has max 25 chars
        add_scope(scope, account)
        mktree(scope, account)
        rule_ids = add_rule(dids=[{'scope': scope, 'name': 'container1213'}], copies=1, rse_expression=f'{self.rse1}|{self.rse2}|{self.rse3}|{self.rse4}',
                            grouping='DATASET', account=account, weight=None, lifetime=None, locked=False, subscription_id=None)
        rule = get_rule(rule_ids[0])
        print(rule['locks_ok_cnt'], rule['locks_replicating_cnt'])
        assert (rule['locks_ok_cnt'] == 17)
        assert (rule['locks_replicating_cnt'] == 8)
        dsl1 = list(get_dataset_locks(scope, 'ds1'))
        dsl2 = list(get_dataset_locks(scope, 'ds2'))
        dsl3 = list(get_dataset_locks(scope, 'ds3'))
        print(dsl1)
        print(dsl2)
        print(dsl3)
        assert (len(dsl1) == 1 and dsl1[0]['rse'] == self.rse1)
        assert (len(dsl2) == 1 and dsl2[0]['rse'] == self.rse3)
        assert (len(dsl3) == 1 and dsl3[0]['rse'] == self.rse1)

        # test3 : NONE grouping
        scope = InternalScope(('scope3_' + str(uuid()))[:21], vo=vo)  # scope field has max 25 chars
        add_scope(scope, account)
        mktree(scope, account)
        rule_ids = add_rule(dids=[{'scope': scope, 'name': 'container1213'}], copies=1, rse_expression=f'{self.rse1}|{self.rse2}|{self.rse3}|{self.rse4}',
                            grouping='NONE', account=account, weight=None, lifetime=None, locked=False, subscription_id=None)
        rule = get_rule(rule_ids[0])
        print(rule['locks_ok_cnt'], rule['locks_replicating_cnt'])
        assert (rule['locks_ok_cnt'] == 25)
        assert (rule['locks_replicating_cnt'] == 0)
        dsl1 = list(get_dataset_locks(scope, 'ds1'))
        dsl2 = list(get_dataset_locks(scope, 'ds2'))
        dsl3 = list(get_dataset_locks(scope, 'ds3'))
        print(dsl1)
        print(dsl2)
        print(dsl3)
        assert (len(dsl1) == 0)
        assert (len(dsl2) == 0)
        assert (len(dsl3) == 0)


def test_rule_boost(vo, mock_scope, rse_factory, jdoe_account):
    """ REPLICATION RULE (CORE): Update a replication rule to quicken the translation from stuck to replicating """
    _, tmp_rse_id = rse_factory.make_mock_rse()
    rse, rse_id = rse_factory.make_mock_rse()
    update_rse(rse_id, {'availability_write': False})
    set_local_account_limit(jdoe_account, rse_id, -1)
    files = create_files(3, mock_scope, tmp_rse_id)
    dataset1 = 'dataset_' + str(uuid())
    add_did(mock_scope, dataset1, DIDType.DATASET, jdoe_account)
    attach_dids(mock_scope, dataset1, files, jdoe_account)

    rule_id = add_rule(dids=[{'scope': mock_scope, 'name': dataset1}], account=jdoe_account, copies=1, rse_expression=rse, grouping='NONE', weight=None, lifetime=None, locked=False, subscription_id=None, ignore_availability=True)[0]
    before_update_rule = {}
    for file in files:
        for filtered_lock in [lock for lock in get_replica_locks(scope=file['scope'], name=file['name'])]:
            assert (filtered_lock['state'] == LockState.STUCK)
            before_update_rule[filtered_lock['name']] = filtered_lock['updated_at']
    before_update_rule_updated_at = get_rule(rule_id)['updated_at']

    update_rule(rule_id, options={'boost_rule': True})

    for file in files:
        for filtered_lock in [lock for lock in get_replica_locks(scope=file['scope'], name=file['name'])]:
            assert (before_update_rule[filtered_lock['name']] > filtered_lock['updated_at'])
    assert (before_update_rule_updated_at > get_rule(rule_id)['updated_at'])


@pytest.mark.usefixtures('setup_class')
class TestClient:

    rule_client = RuleClient()

    def test_add_rule(self, mock_scope, did_factory, jdoe_account):
        """ REPLICATION RULE (CLIENT): Add a replication rule and list full history """
        activity = get_schema_value('ACTIVITY')['enum'][0]
        files = create_files(3, mock_scope, self.rse1_id)
        dataset = did_factory.random_dataset_did()
        add_did(did_type=DIDType.DATASET, account=jdoe_account, **dataset)
        attach_dids(dids=files, account=jdoe_account, **dataset)

        ret = self.rule_client.add_replication_rule(dids=[{'scope': mock_scope.external, 'name': dataset['name']}], account='jdoe', copies=2, rse_expression=self.T1, grouping='NONE', activity=activity)
        assert isinstance(ret, list)

        rep_rules = [rep_rule for rep_rule in self.rule_client.list_replication_rule_full_history(mock_scope.external, dataset['name'])]
        assert len(rep_rules) == 1
        assert ret[0] == rep_rules[0]['rule_id']

    def test_delete_rule(self, mock_scope, did_factory, jdoe_account):
        """ REPLICATION RULE (CLIENT): Delete a replication rule """
        files = create_files(3, mock_scope, self.rse1_id)
        dataset = did_factory.random_dataset_did()
        add_did(did_type=DIDType.DATASET, account=jdoe_account, **dataset)
        attach_dids(dids=files, account=jdoe_account, **dataset)

        rule_id = add_rule(dids=[dataset], account=jdoe_account, copies=1, rse_expression=self.rse1, grouping='NONE', weight='fakeweight', lifetime=None, locked=False, subscription_id=None)[0]

        ret = self.rule_client.delete_replication_rule(rule_id=rule_id)
        assert (ret is True)
        get = self.rule_client.get_replication_rule(rule_id)
        assert (get['expires_at'] is not None)

    def test_list_rules_by_did(self, mock_scope, did_factory, jdoe_account, rucio_client):
        """ DID (CLIENT): List Replication Rules per DID """
        files = create_files(3, mock_scope, self.rse1_id)
        dataset = did_factory.random_dataset_did()
        add_did(did_type=DIDType.DATASET, account=jdoe_account, **dataset)
        attach_dids(dids=files, account=jdoe_account, **dataset)

        rule_id_1 = add_rule(dids=[dataset], account=jdoe_account, copies=1, rse_expression=self.rse1, grouping='NONE', weight='fakeweight', lifetime=None, locked=False, subscription_id=None)[0]

        rule_id_2 = add_rule(dids=[dataset], account=jdoe_account, copies=1, rse_expression=self.rse2, grouping='NONE', weight='fakeweight', lifetime=None, locked=False, subscription_id=None)[0]

        ret = rucio_client.list_did_rules(scope=mock_scope.external, name=dataset['name'])
        ids = [rule['id'] for rule in ret]

        assert rule_id_1 in ids
        assert rule_id_2 in ids

    def test_list_rules_by_name(self, mock_scope, did_factory, jdoe_account, rucio_client):
        """ NAME (CLIENT): List Replication Rules per NAME """
        files = create_files(1, mock_scope, self.rse1_id)
        dataset = did_factory.random_dataset_did()
        add_did(did_type=DIDType.DATASET, account=jdoe_account, **dataset)
        attach_dids(dids=files, account=jdoe_account, **dataset)

        rule_id_1 = add_rule(dids=[dataset], account=jdoe_account, copies=1, rse_expression=self.rse1, grouping='NONE', weight='fakeweight', lifetime=None, locked=False, subscription_id=None)[0]
        rule_id_2 = add_rule(dids=[dataset], account=jdoe_account, copies=1, rse_expression=self.rse2, grouping='NONE', weight='fakeweight', lifetime=None, locked=False, subscription_id=None)[0]
        ret = rucio_client.list_associated_rules_for_file(scope=mock_scope.external, name=files[0]['name'])
        ids = [rule['id'] for rule in ret]

        assert rule_id_1 in ids
        assert rule_id_2 in ids

    def test_get_rule(self, mock_scope, did_factory, jdoe_account):
        """ REPLICATION RULE (CLIENT): Get Replication Rule by id """
        activity = get_schema_value('ACTIVITY')['enum'][0]
        files = create_files(3, mock_scope, self.rse1_id)
        dataset = did_factory.random_dataset_did()
        add_did(did_type=DIDType.DATASET, account=jdoe_account, **dataset)
        attach_dids(dids=files, account=jdoe_account, **dataset)

        ret = self.rule_client.add_replication_rule(dids=[{'scope': mock_scope.external, 'name': dataset['name']}], account='jdoe', copies=2, rse_expression=self.T1, grouping='NONE', activity=activity)
        get = self.rule_client.get_replication_rule(ret[0])
        assert (ret[0] == get['id'])

    def test_get_rule_by_account(self, mock_scope, did_factory, jdoe_account, rucio_client):
        """ ACCOUNT (CLIENT): Get Replication Rule by account """
        activity = get_schema_value('ACTIVITY')['enum'][0]
        files = create_files(3, mock_scope, self.rse1_id)
        dataset = did_factory.random_dataset_did()
        add_did(did_type=DIDType.DATASET, account=jdoe_account, **dataset)
        attach_dids(dids=files, account=jdoe_account, **dataset)

        ret = self.rule_client.add_replication_rule(dids=[{'scope': mock_scope.external, 'name': dataset['name']}], account='jdoe', copies=2, rse_expression=self.T1, grouping='NONE', activity=activity)
        get = rucio_client.list_account_rules('jdoe')
        rules = [rule['id'] for rule in get]

        assert ret[0] in rules

    def test_locked_rule(self, mock_scope, did_factory, jdoe_account):
        """ REPLICATION RULE (CLIENT): Delete a locked replication rule"""
        files = create_files(3, mock_scope, self.rse1_id)
        dataset = did_factory.random_dataset_did()
        add_did(did_type=DIDType.DATASET, account=jdoe_account, **dataset)
        attach_dids(dids=files, account=jdoe_account, **dataset)

        rule_id_1 = add_rule(dids=[dataset], account=jdoe_account, copies=1, rse_expression=self.rse1, grouping='NONE', weight='fakeweight', lifetime=None, locked=True, subscription_id=None)[0]

        pytest.raises(UnsupportedOperation, delete_rule, rule_id_1)
        self.rule_client.update_replication_rule(rule_id=rule_id_1, options={'locked': False})
        delete_rule(rule_id=rule_id_1)

    def test_dataset_lock(self, mock_scope, did_factory, jdoe_account, rucio_client):
        """ DATASETLOCK (CLIENT): Get a datasetlock for a specific dataset"""
        files = create_files(3, mock_scope, self.rse1_id)
        dataset = did_factory.random_dataset_did()
        add_did(did_type=DIDType.DATASET, account=jdoe_account, **dataset)
        attach_dids(dids=files, account=jdoe_account, **dataset)

        rule_id_1 = add_rule(dids=[dataset], account=jdoe_account, copies=1, rse_expression=self.rse1, grouping='DATASET', weight='fakeweight', lifetime=None, locked=True, subscription_id=None)[0]

        rule_ids = [lock['rule_id'] for lock in rucio_client.get_dataset_locks(scope=mock_scope.external, name=dataset['name'])]
        assert rule_id_1 in rule_ids

    def test_change_rule_lifetime(self, mock_scope, did_factory, jdoe_account):
        """ REPLICATION RULE (CLIENT): Change rule lifetime"""
        files = create_files(3, mock_scope, self.rse1_id)
        dataset = did_factory.random_dataset_did()
        add_did(did_type=DIDType.DATASET, account=jdoe_account, **dataset)
        attach_dids(dids=files, account=jdoe_account, **dataset)

        rule_id_1 = add_rule(dids=[dataset], account=jdoe_account, copies=1, rse_expression=self.rse1, grouping='DATASET', weight='fakeweight', lifetime=None, locked=True, subscription_id=None)[0]

        get = self.rule_client.get_replication_rule(rule_id_1)

        self.rule_client.update_replication_rule(rule_id_1, options={'lifetime': -1})

        get2 = self.rule_client.get_replication_rule(rule_id_1)

        assert (get['expires_at'] != get2['expires_at'])

    def test_approve_rule(self, mock_scope, did_factory, jdoe_account):
        """ REPLICATION RULE (CLIENT): Approve rule"""
        files = create_files(3, mock_scope, self.rse1_id)
        dataset = did_factory.random_dataset_did()
        add_did(did_type=DIDType.DATASET, account=jdoe_account, **dataset)
        attach_dids(dids=files, account=jdoe_account, **dataset)

        rule_id = add_rule(dids=[dataset], account=jdoe_account, copies=1, rse_expression=self.rse1, grouping='DATASET', weight='fakeweight', lifetime=None, locked=True, subscription_id=None, ask_approval=True)[0]
        rule = self.rule_client.get_replication_rule(rule_id)
        assert rule['state'] == RuleState.WAITING_APPROVAL.name
        self.rule_client.approve_replication_rule(rule_id)
        rule = self.rule_client.get_replication_rule(rule_id)
        assert rule['state'] == RuleState.INJECT.name


def test_add_rule_with_0_copies(did_client, did_factory, root_account, rse_factory):
    """ REPLICATION RULE (CLIENT): Add a replication rule and list full history """
    rse, rse_id = rse_factory.make_posix_rse()
    file = did_factory.upload_test_file(rse)
    dataset_internal = did_factory.make_dataset()
    container_internal = did_factory.make_container()

    # make all scopes external
    file, dataset, container = ({'scope': did['scope'].external, 'name': did['name']} for did in (file, dataset_internal, container_internal))

    # Attach dataset to container
    did_client.add_files_to_dataset(files=[file], **dataset)
    did_client.add_datasets_to_container(dsns=[dataset], **container)

    with pytest.raises(InvalidValueForKey) as e:
        add_rule(dids=[container_internal], account=root_account, copies=0, rse_expression=rse, grouping='ALL', weight=None, lifetime=None, locked=False, subscription_id=None)
        assert "The number of copies for a replication rule should be greater than 0" in str(e)


@pytest.mark.noparallel(reason='Asynchronos behavior when loading locks')
def test_detach_dataset_lock_removal(did_client, did_factory, root_account, rse_factory, vo):
    rse1, rse1_id = rse_factory.make_posix_rse()
    rse2, rse2_id = rse_factory.make_posix_rse()
    file = did_factory.upload_test_file(rse1)
    dataset_internal = did_factory.make_dataset()
    container_internal = did_factory.make_container()

    # make all scopes external
    file, dataset, container = ({'scope': did['scope'].external, 'name': did['name']} for did in (file, dataset_internal, container_internal))

    # Attach dataset to container
    did_client.add_files_to_dataset(files=[file], **dataset)
    did_client.add_datasets_to_container(dsns=[dataset], **container)

    add_rse_attribute(rse_id=rse1_id, key='fakeweight', value=5)
    add_rse_attribute(rse_id=rse2_id, key='fakeweight', value=5)

    rule_id = add_rule(dids=[container_internal], account=root_account, copies=2, rse_expression='fakeweight>0', grouping='ALL', weight=None, lifetime=None, locked=False, subscription_id=None)[0]
    print("Rule id: {0}".format(rule_id))
    dataset_locks = list(get_dataset_locks(scope=dataset_internal['scope'], name=dataset['name']))
    print("Dataset locks before detach: {0}".format(dataset_locks))
    assert (len([d for d in dataset_locks if d["rule_id"] == rule_id]) == 2)

    # Detach dataset from container, this should delete all locks on the dataset
    did_client.detach_dids(**container, dids=[dataset_internal])

    re_evaluator(once=True, did_limit=None)

    dataset_locks = list(get_dataset_locks(**dataset_internal))
    print("Dataset locks after detach: {0}".format(dataset_locks))
    assert (len([d for d in dataset_locks if d["rule_id"] == rule_id]) == 0)


@pytest.mark.noparallel(reason='Asynchronos behavior when loading locks')
def test_detach_dataset_lock_removal_shared_dataset(did_client, did_factory, root_account, rse_factory, vo):
    rse1, rse1_id = rse_factory.make_posix_rse()
    rse2, rse2_id = rse_factory.make_posix_rse()
    file = did_factory.upload_test_file(rse1)
    dataset_internal = did_factory.make_dataset()
    container_internal = did_factory.make_container()
    container_internal_2 = did_factory.make_container()

    # make all scopes external
    file, dataset, container, container_2 = ({'scope': did['scope'].external, 'name': did['name']} for did in (file, dataset_internal, container_internal, container_internal_2))

    # Attach dataset to container
    did_client.add_files_to_dataset(files=[file], **dataset)
    did_client.add_datasets_to_container(dsns=[dataset], **container)
    did_client.add_datasets_to_container(dsns=[dataset], **container_2)

    add_rse_attribute(rse_id=rse1_id, key='fakeweight', value=5)
    add_rse_attribute(rse_id=rse2_id, key='fakeweight', value=5)

    rule_id = add_rule(dids=[container_internal], account=root_account, copies=2, rse_expression='fakeweight>0', grouping='ALL', weight=None, lifetime=None, locked=False, subscription_id=None)[0]
    print("Rule id: {0}".format(rule_id))
    dataset_locks = list(get_dataset_locks(scope=dataset_internal['scope'], name=dataset['name']))
    print("Dataset locks before detach: {0}".format(dataset_locks))
    assert (len([d for d in dataset_locks if d["rule_id"] == rule_id]) == 2)

    # Detach dataset from container, this should delete all locks on the dataset
    did_client.detach_dids(**container, dids=[dataset_internal])

    re_evaluator(once=True, did_limit=None)

    dataset_locks = list(get_dataset_locks(**dataset_internal))
    print("Dataset locks after detach: {0}".format(dataset_locks))
    assert (len([d for d in dataset_locks if d["rule_id"] == rule_id]) == 0)


@pytest.mark.noparallel(reason='Asynchronos behavior when loading locks')
def test_detach_dataset_lock_removal_shared_file(did_client, did_factory, root_account, rse_factory, vo):
    rse1, rse1_id = rse_factory.make_posix_rse()
    rse2, rse2_id = rse_factory.make_posix_rse()
    file = did_factory.upload_test_file(rse1)
    dataset_internal = did_factory.make_dataset()
    dataset_internal_2 = did_factory.make_dataset()
    container_internal = did_factory.make_container()

    # make all scopes external
    file, dataset, dataset_2, container = ({'scope': did['scope'].external, 'name': did['name']} for did in (file, dataset_internal, dataset_internal_2, container_internal))

    # Attach dataset to container
    did_client.add_files_to_dataset(files=[file], **dataset)
    did_client.add_files_to_dataset(files=[file], **dataset_2)
    did_client.add_datasets_to_container(dsns=[dataset], **container)

    add_rse_attribute(rse_id=rse1_id, key='fakeweight', value=5)
    add_rse_attribute(rse_id=rse2_id, key='fakeweight', value=5)

    rule_id = add_rule(dids=[container_internal], account=root_account, copies=2, rse_expression='fakeweight>0', grouping='ALL', weight=None, lifetime=None, locked=False, subscription_id=None)[0]
    print("Rule id: {0}".format(rule_id))
    dataset_locks = list(get_dataset_locks(scope=dataset_internal['scope'], name=dataset['name']))
    print("Dataset locks before detach: {0}".format(dataset_locks))
    assert (len([d for d in dataset_locks if d["rule_id"] == rule_id]) == 2)

    # Detach dataset from container, this should delete all locks on the dataset
    did_client.detach_dids(**container, dids=[dataset_internal])

    re_evaluator(once=True, did_limit=None)

    dataset_locks = list(get_dataset_locks(**dataset_internal))
    print("Dataset locks after detach: {0}".format(dataset_locks))
    assert (len([d for d in dataset_locks if d["rule_id"] == rule_id]) == 0)


def test_move_rule_invalid_argument(vo, rse_factory, did_factory, mock_scope, root_account):
    rse, rse_id = rse_factory.make_rse()
    new_rse, new_rse_id = rse_factory.make_rse()
    files = create_files(3, mock_scope, [rse_id])
    dataset = did_factory.random_dataset_did()
    add_did(did_type=DIDType.DATASET, account=root_account, **dataset)
    attach_dids(dids=files, account=root_account, **dataset)

    rule_id = add_rule(dids=[dataset], account=root_account, copies=1, rse_expression=rse, grouping='DATASET', weight=None, lifetime=None, locked=False, subscription_id=None)[0]
    assert (get_rule(rule_id)['state'] == RuleState.OK)

    with pytest.raises(UnsupportedOperation):
        _ = move_rule(rule_id, new_rse, override={'session': 17})

    with pytest.raises(UnsupportedOperation):
        _ = move_rule(rule_id, new_rse, override={'xX_MyFirstStreetName_Xx': 17})



# Copyright European Organization for Nuclear Research (CERN) since 2012
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
from sqlalchemy import delete

from rucio.common.schema import get_schema_value
from rucio.core.rse import get_rse_usage
from rucio.daemons.abacus.rse import rse_update
from rucio.daemons.judge import cleaner
from rucio.daemons.reaper import reaper
from rucio.db.sqla import models
from rucio.db.sqla.session import get_session


@pytest.mark.noparallel(reason='uses daemon, failing in parallel to other tests')
class TestAbacusRSE:

    def test_abacus_rse(self, vo, mock_scope, rse_factory, did_factory, rucio_client):
        """ ABACUS (RSE): Test update of RSE usage. """
        # Get RSE usage of all sources
        session = get_session()
        for model in [models.UpdatedRSECounter, models.RSEUsage]:
            stmt = delete(model)
            session.execute(stmt)
        session.commit()

        # Upload files -> RSE usage should increase
        file_sizes = 2
        nfiles = 2
        rse, rse_id = rse_factory.make_posix_rse()
        dids = did_factory.upload_test_dataset(rse_name=rse, scope=mock_scope.external, size=file_sizes, nb_files=nfiles)
        files = [{'scope': did['did_scope'], 'name': did['did_name']} for did in dids]
        dataset = dids[0]['dataset_name']
        rse_update(once=True)
        rse_usage = get_rse_usage(rse_id=rse_id)[0]
        assert rse_usage['used'] == len(files) * file_sizes
        rse_usage_from_rucio = get_rse_usage(rse_id=rse_id, source='rucio')[0]
        assert rse_usage_from_rucio['used'] == len(files) * file_sizes
        rse_usage_from_unavailable = get_rse_usage(rse_id=rse_id, source='unavailable')
        assert len(rse_usage_from_unavailable) == 0

        # Delete files -> rse usage should decrease
        from rucio.daemons.reaper.reaper import REGION
        REGION.invalidate()
        activity = get_schema_value('ACTIVITY')['enum'][0]
        rucio_client.add_replication_rule([{'scope': mock_scope.external, 'name': dataset}], 1, rse, lifetime=-1, activity=activity)
        cleaner.run(once=True)
        if vo:
            reaper.run(once=True, include_rses='vo=%s&(%s)' % (str(vo), rse), greedy=True)
        else:
            reaper.run(once=True, include_rses=rse, greedy=True)
        rse_update(once=True)
        rse_usage = get_rse_usage(rse_id=rse_id)[0]
        assert rse_usage['used'] == 0
        rse_usage_from_rucio = get_rse_usage(rse_id=rse_id, source='rucio')[0]
        assert rse_usage_from_rucio['used'] == 0
        rse_usage_from_unavailable = get_rse_usage(rse_id=rse_id, source='unavailable')
        assert len(rse_usage_from_unavailable) == 0



# Copyright European Organization for Nuclear Research (CERN) since 2012
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import click

from rucio.cli.bin_legacy.rucio_admin import add_subscription, list_subscriptions, reevaluate_did_for_subscription, update_subscription
from rucio.cli.utils import Arguments


@click.group()
def subscription():
    "The methods for automated and regular processing of some specific rules"


@subscription.command("show")
@click.option("-a", "--account", help="Account associated with the subscription")
@click.option("--long", default=False, is_flag=True, help="Show extended information about the subscription")
@click.argument("subscription-name")
@click.pass_context
def list_(ctx, subscription_name, account, long):
    """Show the attributes of a subscription [SUBSCRIPTION-NAME]"""
    args = Arguments({"subs_account": account, "name": subscription_name, "long": long})
    list_subscriptions(args, ctx.obj.client, ctx.obj.logger, ctx.obj.console, ctx.obj.spinner)


@subscription.command("update")
@click.argument("subscription-name")
@click.option("--filter", "did_filter", help='Json serializable DID filter (eg \'{"scope": ["tests"], "project": ["data12_8TeV"]}\')', required=True)
@click.option("--rule", help='List of replication rules (eg \'[{"activity": "Functional Tests", "copies": 2, "rse_expression": "tier=2", "lifetime": 3600, "weight": "mou"}]\')', required=True)
@click.option("--comment", help="Comments on subscription")
@click.option("--lifetime", type=int, help="Subscription lifetime (in days)")
@click.option("--account", help="Account name")
@click.option("--priority", help="The priority of the subscription")
@click.pass_context
def update(ctx, subscription_name, did_filter, rule, comment, lifetime, account, priority):
    """Update a subscription [SUBSCRIPTION-NAME] to have new properties"""
    args = Arguments({"name": subscription_name, "filter": did_filter, "replication_rules": rule, "comments": comment, "lifetime": lifetime, "subs_account": account, "priority": priority})
    update_subscription(args, ctx.obj.client, ctx.obj.logger, ctx.obj.console, ctx.obj.spinner)


@subscription.command("add")
@click.argument("subscription-name")
@click.option("--filter", "did_filter", help='Json serializable DID filter (eg \'{"scope": ["tests"], "project": ["data12_8TeV"]}\')', required=True)
@click.option("--rule", help='List of replication rules (eg \'[{"activity": "Functional Tests", "copies": 2, "rse_expression": "tier=2", "lifetime": 3600, "weight": "mou"}]\')', required=True)
@click.option("--comment", help="Comments on subscription")
@click.option("--lifetime", type=int, help="Subscription lifetime (in days)")
@click.option("--account", help="Account name")
@click.option("--priority", help="The priority of the subscription")
@click.pass_context
def add_(ctx, subscription_name, did_filter, rule, comment, lifetime, account, priority):
    """Create a new subscription with the name [SUBSCRIPTION-NAME]"""
    args = Arguments({"name": subscription_name, "filter": did_filter, "replication_rules": rule, "comments": comment, "lifetime": lifetime, "subs_account": account, "priority": priority})
    add_subscription(args, ctx.obj.client, ctx.obj.logger, ctx.obj.console, ctx.obj.spinner)


@subscription.command("touch")
@click.argument("dids", nargs=-1)
@click.pass_context
def touch(ctx, dids):
    """Reevaluate list of DIDs against all active subscriptions"""
    # TODO make reeval accept dids as a list
    dids = ",".join(dids)
    reevaluate_did_for_subscription(Arguments({"dids": dids}), ctx.obj.client, ctx.obj.logger, ctx.obj.console, ctx.obj.spinner)



# Copyright European Organization for Nuclear Research (CERN) since 2012
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import itertools
import json
import logging
import re
from configparser import NoOptionError, NoSectionError
from functools import wraps
from time import time
from typing import TYPE_CHECKING, Any, Literal, Optional, TypeVar, Union

import flask
from flask.views import MethodView
from werkzeug.datastructures import Headers
from werkzeug.exceptions import HTTPException
from werkzeug.wrappers import Request, Response

from rucio.common import config
from rucio.common.exception import CannotAuthenticate, DatabaseException, IdentityError, RucioException, UnsupportedRequestedContentType
from rucio.common.schema import get_schema_value
from rucio.common.utils import generate_uuid, render_json
from rucio.core.vo import map_vo
from rucio.gateway.authentication import validate_auth_token
from rucio.gateway.identity import get_default_account, list_accounts_for_identity, verify_identity

if TYPE_CHECKING:
    from collections.abc import Callable, Iterable

    from _typeshed import SupportsIter
    from _typeshed.wsgi import StartResponse, WSGIApplication, WSGIEnvironment
    from flask.typing import ResponseReturnValue

    from rucio.web.rest.flaskapi.v1.types import HeadersType

ResponseTypeVar = TypeVar('ResponseTypeVar', bound=flask.wrappers.Response)


class CORSMiddleware:
    """
    WebUI 2.0 makes preflight requests to the API, which are not handled by the API.
    This middleware intercepts the preflight OPTIONS requests and returns a 200 OK response.
    """

    def __init__(self, app: 'WSGIApplication') -> None:
        self.app = app

    def __call__(self, environ: 'WSGIEnvironment', start_response: 'StartResponse') -> 'Iterable[bytes]':
        request: Request = Request(environ)

        if request.environ.get('REQUEST_METHOD') == 'OPTIONS':
            try:
                webui_urls = config.config_get_list('webui', 'urls')
            except (NoOptionError, NoSectionError, RuntimeError) as error:
                logging.exception('Could not get webui urls from config file')
                return str(error), 500  # type: ignore (return type incompatible with Flask middleware)
            if request.origin in webui_urls:
                response: Response = Response(status=200)
                response.headers['Access-Control-Allow-Origin'] = request.origin
                response.headers['Access-Control-Allow-Methods'] = 'GET, POST, PUT, DELETE, OPTIONS'
                response.headers['Access-Control-Allow-Headers'] = request.environ.get('HTTP_ACCESS_CONTROL_REQUEST_HEADERS')  # type: ignore (value could be None)
                response.headers['Access-Control-Allow-Credentials'] = 'true'
                return response(environ, start_response)
            response: Response = Response(status=403)
            return response(environ, start_response)

        # bypass this middleware for non-OPTIONS requests
        return self.app(environ, start_response)


class ErrorHandlingMethodView(MethodView):
    """
    Special MethodView that handles generic RucioExceptions and more generic
    Exceptions for all defined methods automatically.
    """

    def get_headers(self) -> Optional['HeadersType']:
        """Can be overridden to add headers to generic error responses."""
        return None

    def dispatch_request(self, *args, **kwargs) -> Union['ResponseReturnValue', flask.wrappers.Response]:
        headers = self.get_headers() or None
        try:
            return super(ErrorHandlingMethodView, self).dispatch_request(*args, **kwargs)
        except HTTPException:
            raise
        except DatabaseException as error:
            if 'QueuePool' in str(error):
                msg = f'DatabaseException in {self.__class__.__module__} {self.__class__.__name__} {flask.request.method}'
                # logged, because this should be the __exception__
                logging.debug(msg, exc_info=True)
                return generate_http_error_flask(
                    status_code=503,
                    exc=error.__class__.__name__,
                    exc_msg=('Currently there are too many requests for the Rucio '
                             'servers to handle. Please try again in a few minutes.'),
                    headers=headers
                )
            else:
                msg = f'DatabaseException in {self.__class__.__module__} {self.__class__.__name__} {flask.request.method}'
                logging.debug(msg, exc_info=True)
                return generate_http_error_flask(
                    status_code=500,
                    exc=error.__class__.__name__,
                    exc_msg='An unknown Database Exception has occurred.',
                    headers=headers
                )

        except RucioException as error:
            # should be caught in the flask view and generate_http_error_flask with a proper HTTP status code returned
            msg = f'Uncaught RucioException in {self.__class__.__module__} {self.__class__.__name__} {flask.request.method}'
            # logged, because this should be the __exception__
            logging.debug(msg, exc_info=True)
            return generate_http_error_flask(
                status_code=500,
                exc=error.__class__.__name__,
                exc_msg=error.args[0],
                headers=headers
            )
        except Exception as error:
            # logged, because this means a programming error
            logging.exception("Internal Error")
            if headers:
                return str(error), 500, headers
            else:
                return str(error), 500


def request_auth_env() -> Optional['ResponseReturnValue']:
    if flask.request.environ.get('REQUEST_METHOD') == 'OPTIONS':
        return '', 200

    auth_token = flask.request.headers.get('X-Rucio-Auth-Token', default=None)

    try:
        auth = validate_auth_token(auth_token)
    except CannotAuthenticate:
        return generate_http_error_flask(401, CannotAuthenticate.__name__, 'Cannot authenticate with given credentials')
    except RucioException as error:
        return generate_http_error_flask(500, error.__class__.__name__, error.args[0])
    except Exception:
        logging.exception('Internal error in validate_auth_token')
        return 'Internal Error', 500

    flask.request.environ['vo'] = auth.get('vo', 'def')
    flask.request.environ['issuer'] = auth.get('account')
    flask.request.environ['identity'] = auth.get('identity')
    flask.request.environ['request_id'] = generate_uuid()
    flask.request.environ['start_time'] = time()


def response_headers(response: ResponseTypeVar) -> ResponseTypeVar:
    response.headers['Access-Control-Allow-Origin'] = flask.request.environ.get('HTTP_ORIGIN')  # type: ignore (value could be None)
    response.headers['Access-Control-Allow-Headers'] = flask.request.environ.get('HTTP_ACCESS_CONTROL_REQUEST_HEADERS')  # type: ignore (value could be None)
    response.headers['Access-Control-Allow-Methods'] = '*'
    response.headers['Access-Control-Allow-Credentials'] = 'true'

    if flask.request.environ.get('REQUEST_METHOD') == 'GET':
        response.headers['Cache-Control'] = 'no-cache, no-store, max-age=0, must-revalidate'
        response.headers['Cache-Control'] = 'post-check=0, pre-check=0'
        response.headers['Pragma'] = 'no-cache'

    return response


def check_accept_header_wrapper_flask(supported_content_types: 'Iterable[str]'):
    """ Decorator to check if an endpoint supports the requested content type. """

    def wrapper(f):
        @wraps(f)
        def decorated(*args, **kwargs):
            if not flask.request.accept_mimetypes.provided:
                # accept anything, if Accept header is not provided
                return f(*args, **kwargs)

            for supported in supported_content_types:
                if supported in flask.request.accept_mimetypes:
                    return f(*args, **kwargs)

            # none matched..
            return generate_http_error_flask(
                status_code=406,
                exc=UnsupportedRequestedContentType.__name__,
                exc_msg=f'The requested content type {flask.request.environ.get("HTTP_ACCEPT")} is not supported. Use {supported_content_types}.'
            )

        return decorated

    return wrapper


def parse_scope_name(scope_name: str, vo: Optional[str]) -> tuple[str, ...]:
    """
    Parses the given scope_name according to the schema's
    SCOPE_NAME_REGEXP and returns a (scope, name) tuple.

    :param scope_name: the scope_name string to be parsed.
    :param vo: the vo currently in use.
    :raises ValueError: when scope_name could not be parsed.
    :returns: a (scope, name) tuple.
    """
    if not vo:
        vo = 'def'

    # why again does that regex start with a slash?
    scope_regex = re.match(get_schema_value('SCOPE_NAME_REGEXP', vo), '/' + scope_name)
    if scope_regex is None:
        raise ValueError('cannot parse scope and name')
    return scope_regex.group(1, 2)


def try_stream(
        generator: 'SupportsIter',
        content_type: Optional[str] = None
) -> flask.Response:
    """
    Peeks at the first element of the passed generator and raises
    an error, if yielding raises. Otherwise returns
    a flask.Response object.

    :param generator: a generator function or an iterator.
    :param content_type: the response's Content-Type.
                         'application/x-json-stream' by default.
    :returns: a response object with the specified Content-Type.
    """
    if not content_type:
        content_type = 'application/x-json-stream'

    it = iter(generator)
    try:
        peek = next(it)
        return flask.Response(flask.stream_with_context(itertools.chain((peek,), it)), content_type=content_type)
    except StopIteration:
        return flask.Response('', content_type=content_type)


def error_headers(exc_cls: str, exc_msg: str) -> dict[str, str]:
    def strip_newlines(msg: str) -> str:
        return msg.replace('\n', ' ').replace('\r', ' ')

    if exc_msg:
        exc_msg = strip_newlines(exc_msg)
        # Truncate too long exc_msg
        oldlen = len(exc_msg)
        exc_msg = exc_msg[:min(oldlen, 125)]
        if len(exc_msg) != oldlen:
            exc_msg = exc_msg + '...'
    return {
        'ExceptionClass': strip_newlines(exc_cls),
        'ExceptionMessage': exc_msg
    }


def _error_response(exc_cls: str, exc_msg: str) -> tuple[dict[str, str], dict[str, str]]:
    data = {'ExceptionClass': exc_cls,
            'ExceptionMessage': exc_msg}
    headers = {'Content-Type': 'application/octet-stream'}
    headers.update(error_headers(exc_cls=exc_cls, exc_msg=exc_msg))
    return data, headers


def generate_http_error_flask(
        status_code: int,
        exc: Union[str, BaseException],
        exc_msg: Optional[str] = None,
        headers: Optional['HeadersType'] = None,
) -> "flask.Response":
    """Utility function to generate a complete HTTP error response.

    :param status_code: The HTTP status code to generate a response for.
    :param exc: The name of the exception class or a RucioException object.
    :param exc_msg: The error message.
    :param headers: any default headers to send along.
    :returns: a response object representing the error.
    """
    if isinstance(exc, BaseException):
        if not exc_msg and exc.args and exc.args[0]:
            exc_msg = exc.args[0]
        exc_cls = exc.__class__.__name__
    else:
        exc_cls = str(exc)
    exc_msg = str(exc_msg)

    data, prioheaders = _error_response(exc_cls, exc_msg)
    headers = Headers(headers)
    headers.extend(prioheaders)
    try:
        return flask.Response(
            status=status_code,
            headers=headers,
            content_type=prioheaders['Content-Type'],
            response=render_json(**data),
        )
    except Exception:
        logging.exception(f'Cannot create generate_http_error_flask response with {data}')
        raise


def json_parameters(json_loads: "Callable[[str], Any]" = json.loads, optional: bool = False) -> dict:
    """
    Returns the JSON parameters from the current request's body as dict.
    """
    if optional:
        kwargs = {'default': {}}
    else:
        kwargs = {}
    return json_parse(types=(dict, ), json_loads=json_loads, **kwargs)


def json_list(json_loads: "Callable[[str], Any]" = json.loads, optional: bool = False) -> list:
    """
    Returns the JSON array from the current request's body as list.
    """
    if optional:
        kwargs = {'default': []}
    else:
        kwargs = {}
    return json_parse(types=(list, ), json_loads=json_loads, **kwargs)


def json_parse(types: tuple, json_loads: "Callable[[str], Any]" = json.loads, **kwargs):
    def clstostr(cls) -> str:
        if cls.__name__ == "dict":
            return "dictionary"
        else:
            return cls.__name__

    def typestostr(_types: tuple) -> str:
        return " or ".join(map(clstostr, _types))

    data = flask.request.get_data(as_text=True)
    if 'default' in kwargs and not data:
        return kwargs['default']
    try:
        body = json_loads(data)
        if not isinstance(body, types):
            flask.abort(
                generate_http_error_flask(
                    status_code=400,
                    exc=TypeError.__name__,
                    exc_msg='body must be a json ' + typestostr(types)
                )
            )
        return body
    except json.JSONDecodeError:
        flask.abort(
            generate_http_error_flask(
                status_code=400,
                exc=ValueError.__name__,
                exc_msg='cannot decode json parameter ' + typestostr(types)
            )
        )


def param_get(parameters: dict[str, Any], name: str, **kwargs) -> Any:
    if 'default' in kwargs:
        return parameters.get(name, kwargs['default'])
    else:
        if name not in parameters:
            flask.abort(
                generate_http_error_flask(
                    status_code=400,
                    exc=KeyError.__name__,
                    exc_msg=f"'{name}' not defined"
                )
            )
        return parameters[name]


def extract_vo(headers: Headers) -> str:
    """ Extract the VO name from the given request.headers object and
        does any name mapping. Returns the short VO name or raise a
        flask.abort if the VO name doesn't meet the name specification.

    :papam headers: The request.headers object for the current request.
    :returns: a string containing the short VO name.
    """
    try:
        return map_vo(headers.get('X-Rucio-VO', default='def'))
    except RucioException as err:
        # VO Name doesn't match allowed spec
        flask.abort(generate_http_error_flask(status_code=400, exc=err))


def get_account_from_verified_identity(
        identity_key: str,
        id_type: Literal["USERPASS", "X509"],
        password: Optional[str] = None
) -> list[str]:
    """ Verifies the provided identity and tries to return a matching account.
        If no account is found, raises an IdentityError after trying to verify the identity.
        If multiple accounts are found, returns the default account if available, otherwise all accounts.
    :param identity_key: The identity key name. For example x509 DN, or a username.
    :param id_type: The type of the authentication (x509, USERPASS).
    :param password: required only if id_type==USERPASS.
    :raises IdentityError: if no account is found for the identity or if the identity could not be verified.
    :returns: a list of account names.
    """
    accounts = list_accounts_for_identity(identity_key=identity_key, id_type=id_type)
    if accounts is None or len(accounts) == 0:
        if id_type == 'USERPASS':
            verify_identity(identity_key=identity_key, id_type=id_type, password=password)
        elif id_type == 'X509':
            verify_identity(identity_key=identity_key, id_type=id_type)
        else:
            raise IdentityError('No account found for identity')
    if len(accounts) > 1:
        try:
            default_account = get_default_account(identity_key=identity_key, id_type=id_type)
            return [default_account]
        except IdentityError:
            return accounts
    else:
        account = accounts[0]
        return [account]



# Copyright European Organization for Nuclear Research (CERN) since 2012
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import importlib
import signal
import time

import click
from rich.console import Console
from rich.status import Status
from rich.theme import Theme
from rich.traceback import install

from rucio import version
from rucio.cli.bin_legacy.rucio import ping, test_server, whoami_account
from rucio.cli.utils import Arguments, exception_handler, get_client, setup_gfal2_logger, signal_handler
from rucio.client.richclient import MAX_TRACEBACK_WIDTH, MIN_CONSOLE_WIDTH, CLITheme, get_cli_config, get_pager, setup_rich_logger
from rucio.common.utils import setup_logger


# Taken directly from https://click.palletsprojects.com/en/stable/complex/#defining-the-lazy-group
class LazyGroup(click.Group):
    def __init__(self, *args, lazy_subcommands=None, **kwargs):
        super().__init__(*args, **kwargs)
        # lazy_subcommands is a map of the form:
        #
        #   {command-name} -> {module-name}.{command-object-name}
        #
        self.lazy_subcommands = lazy_subcommands or {}

    def list_commands(self, ctx):
        base = super().list_commands(ctx)
        lazy = sorted(self.lazy_subcommands.keys())
        return base + lazy

    def get_command(self, ctx, cmd_name):
        if cmd_name in self.lazy_subcommands:
            return self._lazy_load(cmd_name)
        return super().get_command(ctx, cmd_name)

    def _lazy_load(self, cmd_name):
        # lazily loading a command, first get the module name and attribute name
        import_path = self.lazy_subcommands[cmd_name]
        modname, cmd_object_name = import_path.rsplit(".", 1)
        # do the import
        mod = importlib.import_module(modname)
        # get the Command object from that module
        cmd_object = getattr(mod, cmd_object_name)
        # check the result to make debugging easier
        if not isinstance(cmd_object, click.BaseCommand):
            raise ValueError(f"Lazy loading of {import_path} failed by returning " "a non-command object")
        return cmd_object


@click.group(
    cls=LazyGroup,
    lazy_subcommands={
        "account": "rucio.cli.account.account",
        "config": "rucio.cli.config.config",
        "did": "rucio.cli.did.did",
        "download": "rucio.cli.download.download",
        "lifetime-exception": "rucio.cli.lifetime_exception.lifetime_exception",
        "replica": "rucio.cli.replica.replica",
        "rse": "rucio.cli.rse.rse",
        "rule": "rucio.cli.rule.rule",
        "scope": "rucio.cli.scope.scope",
        "subscription": "rucio.cli.subscription.subscription",
        "upload": "rucio.cli.upload.upload_command",
    },
    context_settings={"help_option_names": ["-h", "--help"]}
)  # TODO: Implement https://click.palletsprojects.com/en/stable/options/#dynamic-defaults-for-prompts for args from config or os
@click.option("--account", "--issuer", "issuer", help="Rucio account to use.")
@click.option("--auth-host", help="The Rucio Authentication host")
@click.option("-S", "--auth-strategy", help="Authentication strategy", type=click.Choice(['userpass', 'x509', 'x509_proxy', 'gss', 'ssh', 'saml', 'oidc']))
# x509 and x509 proxy auth
@click.option("--ca-certificate", help="CA certificate to verify peer against (SSL)")
@click.option("--certificate", help="Client certificate file")
@click.option("--client-key", help="Client key for x509 Authentication")
@click.option("--config", help="The Rucio configuration file to use")
@click.option("-H", "--host", help="The Rucio API host")
# oidc auth
@click.option("--oidc-user", help="OIDC username")
@click.option("--oidc-password", help="OIDC password")
@click.option("--oidc-audience", help="Defines which audience are tokens requested for.")
@click.option(
    "--oidc-auto",
    is_flag=True,
    default=False,
    help="""
        If not specified, username and password credentials are not required and users will be given a URL to use in their browser.
        If specified, the users explicitly trust Rucio with their IdP credentials"
    """,
)
@click.option(
    "--oidc-issuer",
    help="""
        Defines which Identity Provider is going to be used.
        The issuer string must correspond to the keys configured in the /etc/idpsecrets.json auth server configuration file")
    """,
)
@click.option(
    "--oidc-polling",
    is_flag=True,
    default=False,
    help="""
        If not specified, user will be asked to enter a code returned by the browser to the command line.
        If --polling is set, Rucio Client should get the token without any further interaction of the user.
        This option is active only if --auto is *not* specified
    """,
)
@click.option(
    "--oidc-refresh-lifetime",
    help="""
        Max lifetime in hours for this access token; the token will be refreshed by an asynchronous Rucio daemon.
        If not specified, refresh will be stopped after 4 days.
        This option is effective only if --oidc-scope includes offline_access scope for a refresh token to be granted to Rucio
    """,
)
@click.option(
    "--oidc-scope",
    default="openid profile",
    help="""
        Defines which (OIDC) information user will share with Rucio. Rucio requires at least -sc='openid profile'.
        To request refresh token for Rucio, scope must include 'openid offline_access'
        and there must be no active access token saved on the side of the currently used Rucio Client,
    """,
)
@click.option("-T", "--timeout", type=float, help="Set all timeout values to seconds")
@click.option("-U", "--user-agent", default="rucio-clients", help="Rucio User Agent")
# userpass/gss/saml auth
@click.option("-u", "--user", help="Username for userpass")
@click.option("--password", help="Password for userpass")
@click.option("--vo", help="VO to authenticate at. Only used in multi-VO mode")
@click.option("-v", "--verbose", default=False, is_flag=True, help="Print more verbose output")
@click.version_option(version.version_string(), message="%(prog)s %(version)s")
# Hidden options at the end
@click.option("--pager", is_flag=True, default=False, hidden=True)
@exception_handler
@click.pass_context
def main(
    ctx,
    config,
    verbose,
    host,
    auth_host,
    issuer,
    auth_strategy,
    timeout,
    user_agent,
    vo,
    pager,
    user,
    password,
    oidc_user,
    oidc_password,
    oidc_scope,
    oidc_audience,
    oidc_auto,
    oidc_polling,
    oidc_refresh_lifetime,
    oidc_issuer,
    certificate,
    client_key,
    ca_certificate,
):
    ctx.ensure_object(Arguments)
    ctx.obj.start_time = time.time()

    ctx.obj.no_pager = not pager
    pager = get_pager()
    use_rich = get_cli_config() == "rich"

    console = Console(theme=Theme(CLITheme.LOG_THEMES), soft_wrap=True)
    console.width = max(MIN_CONSOLE_WIDTH, console.width)
    spinner = Status("Initializing spinner", spinner=CLITheme.SPINNER, spinner_style=CLITheme.SPINNER_STYLE, console=console)

    ctx.obj.use_rich = use_rich
    ctx.obj.spinner = spinner
    ctx.obj.console = console
    ctx.obj.pager = pager

    if use_rich:
        install(console=console, word_wrap=True, width=min(console.width, MAX_TRACEBACK_WIDTH))  # Make rich exception tracebacks the default.
        logger = setup_rich_logger(module_name=__name__, logger_name="user", verbose=verbose, console=console)
    else:
        logger = setup_logger(module_name=__name__, logger_name="user", verbose=verbose)
    args = Arguments(
        {
            "config": config,
            "host": host,
            "isser": issuer,
            "auth_host": auth_host,
            "auth_strategy": auth_strategy,
            "timeout": timeout,
            "user_agent": user_agent,
            "VO": vo,
            "username": user,
            "password": password,
            "oidc_username": oidc_user,
            "oidc_password": oidc_password,
            "oidc_scope": oidc_scope,
            "oidc_audience": oidc_audience,
            "oidc_auto": oidc_auto,
            "oidc_polling": oidc_polling,
            "oidc_refresh_lifetime": oidc_refresh_lifetime,
            "oidc_issuer": oidc_issuer,
            "certificate": certificate,
            "client_key": client_key,
            "ca_certificate": ca_certificate,
        }
    )  # TODO Future improvement - change `get_client` to take these args directly
    client = get_client(args, logger)  # TODO Future improvement - use envvar functionality in click to remove conditionals checking env vars

    setup_gfal2_logger()
    signal.signal(signal.SIGINT, lambda sig, frame: signal_handler(sig, frame, logger))

    ctx.obj.client = client
    ctx.obj.logger = logger

    ctx.call_on_close(_teardown)


@click.pass_context
def _teardown(ctx):
    time_elapsed = time.time() - ctx.obj.start_time

    if ctx.obj.use_rich:
        ctx.obj.spinner.stop()
    if ctx.obj.console.is_terminal and not ctx.obj.no_pager:
        command_output = ctx.obj.console.end_capture()
        if command_output == "" and ctx.obj.verbose:
            print("Completed in %-0.4f sec." % (time_elapsed))
        else:
            if ctx.obj.verbose:
                command_output += "Completed in %-0.4f sec." % (time_elapsed)
            # Ignore SIGINT during pager execution.
            signal.signal(signal.SIGINT, signal.SIG_IGN)
            ctx.obj.pager(command_output)
    else:
        if ctx.obj.verbose:
            print("Completed in %-0.4f sec." % (time_elapsed))


@main.command(name="whoami", help="Get information about account whose token is used")
@click.pass_context
def exe_whoami(ctx):
    args = Arguments()
    whoami_account(args, ctx.obj.client, ctx.obj.logger, ctx.obj.console, ctx.obj.spinner)


@main.command(name="ping", help="Ping Rucio server")
@click.pass_context
def exe_ping(ctx):
    args = Arguments()
    ping(args, ctx.obj.client, ctx.obj.logger, ctx.obj.console, ctx.obj.spinner)


@main.command(name="test-server", help="Test client against the server")
@click.pass_context
def exe_test_server(ctx):
    args = Arguments()
    test_server(args, ctx.obj.client, ctx.obj.logger, ctx.obj.console, ctx.obj.spinner)



# Copyright European Organization for Nuclear Research (CERN) since 2012
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

''' add a new state LOST in BadFilesStatus '''

from alembic import context
from alembic.op import create_check_constraint

from rucio.db.sqla.util import try_drop_constraint

# Alembic revision identifiers
revision = '3d9813fab443'
down_revision = '1fc15ab60d43'


def upgrade():
    '''
    Upgrade the database to this revision
    '''

    if context.get_context().dialect.name in ['oracle', 'mysql', 'postgresql']:
        create_check_constraint(constraint_name='BAD_REPLICAS_STATE_CHK', table_name='bad_replicas',
                                condition="state in ('B', 'D', 'L', 'R', 'S')")


def downgrade():

    '''
    Downgrade the database to the previous revision
    '''

    if context.get_context().dialect.name in ['oracle', 'postgresql']:
        try_drop_constraint('BAD_REPLICAS_STATE_CHK', 'bad_replicas')



# Copyright European Organization for Nuclear Research (CERN) since 2012
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from io import StringIO

from rucio.common import dumper
from rucio.tests.common import mock_open

from .mocks import gfal2


def test_gfal_download_to_file():
    gfal_file = StringIO()
    local_file = StringIO()
    gfal_file.write('content')
    gfal_file.seek(0)

    with gfal2.mocked_gfal2(dumper, files={'srm://example.com/file': gfal_file}):
        dumper.gfal_download_to_file('srm://example.com/file', local_file)

    local_file.seek(0)
    assert local_file.read() == 'content'


def test_gfal_download_creates_file_with_content():
    gfal_file = StringIO()
    local_file = StringIO()
    gfal_file.write('content')
    gfal_file.seek(0)

    with gfal2.mocked_gfal2(dumper, files={'srm://example.com/file': gfal_file}):
        with mock_open(dumper, local_file):
            dumper.gfal_download('srm://example.com/file', 'filename')

    local_file.seek(0)
    assert local_file.read() == 'content'



# Copyright European Organization for Nuclear Research (CERN) since 2012
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import click

from rucio.cli.bin_legacy.rucio_admin import delete_config_option, get_config, set_config_option
from rucio.cli.utils import Arguments


@click.group()
def config():
    "Modify the configuration table"


# TODO Limit to just the section names
@config.command("list")
@click.option("-s", "--section", help="Filter by sections")
@click.option("-k", "--key", help="Show key's value, section required.")
@click.pass_context
def list_(ctx, section, key):
    """List the sections or content of sections in the rucio.cfg"""
    get_config(Arguments({"section": section, "key": key}), ctx.obj.client, ctx.obj.logger, ctx.obj.console, ctx.obj.spinner)


# TODO Change to only add new fields and cannot modify an existing field
@config.command("add")
@click.option("-s", "--section", help="Section name", required=True)
@click.option('--key', help='Attribute key', required=True)
@click.option('--value', help='Attribute value', required=True)
@click.pass_context
def add_(ctx, section, key, value):
    """
    Add a new key/value to a section.

    \b
    Example, Add a key to an existing section:
        $ rucio config add --section my-section --key key --value value
    """
    args = Arguments({"section": section, "option": key, "value": value})
    set_config_option(args, ctx.obj.client, ctx.obj.logger, ctx.obj.console, ctx.obj.spinner)


@config.command("remove")
@click.option("-s", "--section", help="Section", required=True)
@click.option("-k", "--key", help="Key in section", required=True)
@click.pass_context
def remove(ctx, section, key):
    """Remove the section.key from the config."""
    args = Arguments({"section": section, "option": key})
    delete_config_option(args, ctx.obj.client, ctx.obj.logger, ctx.obj.console, ctx.obj.spinner)


# @config.command("show")
# @click.pass_context
def show(ctx):
    """Show a single sections options"""


# TODO Change this so that it only modifies existing fields
def update():
    """Modify an existing command"""
    pass
